import argparse
import glob
import sys

# # minimum number of particles required in the halos/galaxies
# min_N = 100
# # only find/return the most massive halos in the catalogues (5 should be
# # sufficient for most cases)
# max_halos = 1
# disc_def = dict() #jzjc_min=0.85, rmax='50 kpc', zmax='10 kpc')

RESTART_FILENAME_postfix = 'restart.gcache3'

parser = argparse.ArgumentParser(
            description='Trace the most massive progenitor of the most massive '
                        '(in stellar mass) galaxy in a highly resolved halo in '
                        'the start snapshot, that has no or very little low '
                        'resolution particles/mass, though the simulation. Store '
                        'its properties as well as storing the corresponding cut '
                        'region (centered at the center of the stars of the '
                        'galaxy of interest using a shrinking sphere and its '
                        'mass-weighted velocity).')
parser.add_argument('name_pattern',
                    help='The pattern of the snapshots\' file names. One ' +
                         'can use C- / Python-like string formatter that ' +
                         'are all going to be replaced by the snapshot ' +
                         'number. (example: "output/snap_%%03d.gdt")')
parser.add_argument('--command', '-c',
                    metavar='str',
                    default='',
                    help='the path+name of the script to process the snapshots (default: "") ')
parser.add_argument('--par',
                    metavar='str',
                    default='',
                    help='parameter to pass to --command (default: "") ')
parser.add_argument('--par1',
                    metavar='str',
                    default='',
                    help='parameter 1 to pass to --command (default: "") ')
parser.add_argument('--par2',
                    metavar='str',
                    default='',
                    help='parameter 2 to pass to --command (default: "") ')
parser.add_argument('--par3',
                    metavar='str',
                    default='',
                    help='parameter 3 to pass to --command (default: "") ')
parser.add_argument('--par4',
                    metavar='str',
                    default='',
                    help='parameter 4 to pass to --command (default: "") ')
parser.add_argument('--par5',
                    metavar='str',
                    default='',
                    help='parameter 5 to pass to --command (default: "") ')
parser.add_argument('--profile', '-p',
                    metavar='str',
                    default='cache',
                    help='the profile used to process the snapshots (default: "cache") ')
parser.add_argument('--withplot',
                    action='store_true',
                    help='create a plot in the cache.')

parser.add_argument('--destination', '-d',
                    metavar='FOLDER',
                    default=None,
                    help='The folder to store the output (default: <profile name> ' +
                         'in the directory of the snapshots).')
parser.add_argument('--start', '-s',
                    metavar='INT',
                    type=int,
                    default=None,
                    help='The number of the snapshot to start with (the ' +
                         'latest one). If there is a file called ' +
                         '"%s" in the destination directory,' % RESTART_FILENAME_postfix +
                         'restart the last run and ignore the start parameter. ' +
                         'By default it is looked for the highest consecutive ' +
                         'number in the snapshot filename pattern for which a ' +
                         'snapshot exists.')
parser.add_argument('--end', '-e',
                    metavar='INT',
                    type=int,
                    default=None,
                    help='The number of the logysnapshot to stop at (the ' +
                         'earliest one); default: %(default)d).')
parser.add_argument('--step',
                    metavar='INT',
                    type=int,
                    default=1,
                    help='The step size in marching the snapshots (default: ' +
                         '%(default)d).')
# moved to environment DEFAULT-values for cache-processing
# parser.add_argument('--linking-length', '-l',
#                     metavar='FLOAT',
#                     type=float,
#                     default=0.05,
#                     help='The linking length used for the FoF finder in terms ' +
#                          'of the mean particle separation. ' +
#                          'Default: %(default)g.')
# parser.add_argument('--linking-vel', '-lv',
#                     metavar='STR',
#                     default='100 km/s',
#                     help='The linking velocity used for the FoF finder. ' +
#                          'Default: %(default)s')
# parser.add_argument('--lowresthreshold',
#                     metavar='FLOAT',
#                     type=float,
#                     default=1e-2,
#                     help='The maximum allowed mass of low resolution elements ' +
#                          'in a halo, defined by a sphere of R200 around the ' +
#                          'center of mass of the FoF group. (default: ' +
#                          '%(default)g).')
# parser.add_argument('--gal-rad',
#                     metavar='FLOAT',
#                     type=float,
#                     default=0.1,
#                     help='The radius in units of R200 to define the galaxy ' +
#                          '(default: %(default)s).')
parser.add_argument('--tlimit',
                    metavar='STR',
                    default='23 h',
                    help='If this time limit in hours is exceeded, stop the ' +
                         'caching (default: %(default)s).')
parser.add_argument('--overwrite',
                    action='store_true',
                    help='Force to overwrite possibly existing files in the ' +
                         'trace folder.')
parser.add_argument('--updatecache',
                    action='store_true',
                    help='Update profile cache at the end of processing ' +
                         'otherwise results are temporary only.')
parser.add_argument('--cachedonly',
                    action='store_true',
                    help='process on snapshots already in cache ' +
                         'others are skipped')
parser.add_argument('--reverseorder',
                    action='store_true',
                    help='process on snapshots from low to high z ')
parser.add_argument('--loadsf',
                    action='store_true',
                    help='load star forming information (insitu)')
# moved to profile properties
# parser.add_argument('--findgxfast',
#                     action='store_true',
#                     help='use fast method to find central galaxy ' +
#                          'otherwise full FoF search is used.')
parser.add_argument('--verbose', '-v',
                    action='store_true',
                    help='Run in verbose mode.')
# writing to snapshot files not supported
# parser.add_argument('--cut',
#                     metavar='FLOAT',
#                     type=float,
#                     default=None,
#                     help='Also save the cut halos (all within <cut>*R200, not ' +
#                          'the FoF group) as Gadget files in the trace folder.')
# parser.add_argument('--length',
#                     default='ckpc/h_0',
#                     help='Define the Gadget length units (default: %(default)s).')
# parser.add_argument('--mass',
#                     default='1e10 Msol h_0**-1',
#                     help='Define the Gadget mass units (default: %(default)s).')
# parser.add_argument('--velocity',
#                     default='km/s',
#                     help='Define the Gadget velocity units (default: %(default)s).')


# def my_center(halo, galaxy, snap, old_center, args):
#     '''The definition of the center of the galaxy/halo.'''
#     if galaxy is not None:
#         center = galaxy.com
#     elif halo is not None:
#         # take all matter within args.gal_rad * R200 like in the definition of the
#         # non-FoF galaxy:
#         R200, M200 = pg.analysis.virial_info(snap, center=old_center, odens=200)
#         inner_halo = snap[pg.BallMask(center=old_center, R=args.gal_rad*R200,
#                                          sph_overlap=False)]
#         center = pg.analysis.center_of_mass(inner_halo)
#         del inner_halo
#     else:
#         center = old_center
#     return center

def get_snap_fname(dirname, pattern, num):
    fparts = os.path.split(pattern)
    nums = (num,) * args.name_pattern.count('%')
    fname = dirname + '/' + fparts[1] % nums
    return fname


def find_last_snapshot(args, dirname):
    '''Find the highest (consecutive) snapshot number.'''
    i_max = int(1e4)
    for num in range(args.end, i_max):
        fparts = os.path.split(args.name_pattern)
        fname = dirname + '/' + fparts[1]
        nums = (num,) * fname.count('%')
        snap_fname = fname % nums
        filelist = glob.glob(snap_fname)
        if len(filelist) > 1:
            print('ERROR: pattern "%s" invalid ' % args.name_pattern,
                  'identifies more than one snapshot!', file=sys.stderr)
            sys.exit(1)
        if len(filelist) == 0:
        #if not os.path.exists(snap_fname):
            if num == args.end:
                snap_fname = os.path.basename(snap_fname)
                print('ERROR: first snapshot "%s"' % snap_fname,
                                     'does not exist!', file=sys.stderr)
                sys.exit(1)
            return num-1

    print('ERROR: did not find last snapshot number in',
                         '[%d,%d] with pattern "%s"!' % (args.end, i_max-1,
                                 os.path.basename(args.name_pattern)), file=sys.stderr)
    sys.exit(1)


def find_first_snapshot(args, dirname):
    '''Find the first snapshot number.'''
    i_max = int(1e4)
    for num in range(0, i_max):
        fparts = os.path.split(args.name_pattern)
        fname = dirname + '/' + fparts[1]
        nums = (num,) * fname.count('%')
        snap_fname = fname % nums
        filelist = glob.glob(snap_fname)
        if len(filelist) > 1:
            print('ERROR: pattern "%s" invalid ' % args.name_pattern,
                  'identifies more than one snapshot!', file=sys.stderr)
            sys.exit(1)
        if len(filelist) == 0:
        #if not os.path.exists(snap_fname):
            if num == i_max:
                snap_fname = os.path.basename(snap_fname)
                print('ERROR: first snapshot "%s"' % snap_fname,
                                     'does not exist!', file=sys.stderr)
                sys.exit(1)
        else:
            return num
    if args.end is None:
        print('ERROR: did not find first snapshot number in',
              '[%d,%d] with pattern "%s"!' % (0, i_max,
                                              args.name_pattern), file=sys.stderr)
    else:
        print('ERROR: did not find first snapshot number in',
                         '[%d,%d] with pattern "%s"!' % (args.end, i_max,
                                 args.name_pattern), file=sys.stderr)
    sys.exit(1)


def load_command(script_file):
    command_str = ''
    if script_file == '':
        return command_str
    filename = script_file
    if len(script_file) > 3 and script_file[:-3].lower() != '.py':
        filename = filename + '.py'
    try:
        with open(filename, 'r') as myfile:
            command_str = ''
            ignore = True
            intent = ''
            for line in myfile:
                if len(line) >= 3 and line[:3] == "def": ignore = False
                if line.find('__name__') >= 0 and line.find('__main__'):
                    ignore = False
                elif len(line) >= 15 and line[:15] == "# gcache3-init:":
                    ignore = True
                elif len(line) >= 18 and line[:18] == "# gcache3-prepare:":
                    command_str += "if snap_exec == 'prepare':\n"
                    line = line.lstrip(' ')
                    intent = '    '
                    ignore = False
                elif len(line) >= 18 and line[:18] == "# gcache3-process:":
                    command_str += "if snap_exec == 'process':\n"
                    line = line.lstrip(' ')
                    intent = '    '
                    ignore = False
                elif len(line) >= 16 and line[:16] == "# gcache3-close:":
                    command_str += "if snap_exec == 'close':\n"
                    line = line.lstrip(' ')
                    intent = '    '
                    ignore = False
                elif len(line) >= 14 and line[:14] == "# gcache3-end:":
                    ignore = True
                elif len(line) >= 1 and line[:1] == "#":
                    line = ''
                if not ignore and line != "":
                    command_str += intent + line
            myfile.close()
    except Exception as e:
        print('*** error loading script ', filename)
        print(e)
        if args.verbose:
            print('ERROR: loading script', filename, file=sys.stderr)
        sys.exit(1)

    return command_str

def plot_snapshot(sf, snapshot, title=None):
    figx = pg.environment.PLOT_figx
    figy = pg.environment.PLOT_figy
    fs = pg.environment.PLOT_fontsize

    prefix = sf.profile
    print("*********************************************")
    print("plotting snapshot ", snapshot.parts)
    print("prefix, title: ", prefix, ", ", title)
    fig, ax = plt.subplots(figsize=(figx, figy))
    if title is not None:
        ax.set_title(title + ' ('+ prefix + ')')

    pg.plotting.image(snapshot.stars, ax=ax)
    fig.legend(fontsize=fs)
    buf = io.BytesIO()
    plt.savefig(buf, format='png')
    plt.close()

    print("plotting complete")
    print("*********************************************")
    return buf


if __name__ == '__main__' or __name__ == 'pygad.cmdtool.gCache': # imported by command line script
    print("gCache arguments specified")
    for cp in sys.argv[1:]:
        print("gCache: argument ", cp)

    args = parser.parse_args()
    # print("gCache arguments in effect")
    # for arg in vars(args):
    #     print(arg, getattr(args, arg))

    import sys
    if args.verbose:
        print('starting up...')
        sys.stdout.flush()

    # Suspend importing to here, after parsing the arguments to be quick,
    # if only the help has to be displayed or there already occurs and
    # error in parsing the arguments.
    import time
    t_start = time.time()
    import os
    import io
    import gc
    from fractions import Fraction
    import numpy as np
    import pickle as pickle
    import pygad as pg                  # to enable easy testing of command-scripts use absolute

    import matplotlib.pyplot as plt

    start_time = time.time()
    if args.verbose:
        pg.environment.verbose = pg.environment.VERBOSE_TACITURN
    else:
        pg.environment.verbose = pg.environment.VERBOSE_QUIET

    # test arguments
    if args.name_pattern.count('%') == 0:
        if args.verbose:
            print('ERROR: the pattern of the snapshot filename',
                                 'does not contain any string formatter. It',
                                 'will be the same name for all snapshots!', file=sys.stderr)
        sys.exit(1)
    # prepare arguments

    basedir = os.getenv('SNAPSHOT_HOME', '')
    if basedir != '' and basedir[-1] != '/': basedir = basedir + '/'

    if args.name_pattern.find('/') > -1:
        dirname = os.path.dirname(basedir + args.name_pattern)
    else:
        dirname = '.'
        basedir = ''
    dirlist = glob.glob(dirname)

    if len(dirlist) == 0:
        print('WARNING: no snapshots directories found - stopping', file=sys.stderr)
        exit(1)

    for snap_dir in dirlist:
        snap_restart_destination = snap_dir + '/' + args.profile

        print('*** process directory ', snap_dir)

        RESTART_FILENAME = snap_restart_destination + '/' + RESTART_FILENAME_postfix
        if args.overwrite and os.path.exists(RESTART_FILENAME):
            os.remove(RESTART_FILENAME)
        if os.path.exists(RESTART_FILENAME):
            if args.verbose:
                print('reading restart file "%s"...' % RESTART_FILENAME)
                print('WARNING: ignoring --start=%s!' % args.start)
                sys.stdout.flush()
            with open(RESTART_FILENAME, 'rb') as f:
                old_args, snap_num = pickle.load(f)
            # check the compability of the arguments
            new_args = vars(args).copy()
            for not_cmp in ['start', 'step', 'command','withplot', 'end', 'tlimit', 'overwrite', 'verbose']:
                old_args.pop(not_cmp)
                new_args.pop(not_cmp)
            # nur zum testen unter kommentar
            # if old_args != new_args:
            #     print('ERROR: arguments are not compatible ' + \
            #                          'with the previous run!', file=sys.stderr)
            #     print('Not matching these old arguments:', file=sys.stderr)
            #     for arg, value in old_args.items():
            #         if arg in new_args and new_args[arg] != value:
            #             print('  %-20s %s' % (arg+':', value), file=sys.stderr)
            #     sys.exit(1)
            del old_args, new_args
            args.start = snap_num-args.step
            if args.verbose:
                print('done reading restart file')
            if args.end is None:
                args.end = find_first_snapshot(args, snap_dir)
            if snap_num == args.end:
                print('WARNING: no snapshot to process last =', snap_num, ' end =', args.end, file=sys.stderr)
                exit(1)

            print('restart: snapshot ', args.start, args.end)

        else:
            ignore_start = False
            if args.end is None:
                args.end = find_first_snapshot(args, snap_dir)
            elif args.end == 999:
                ignore_start = True
                args.end = find_first_snapshot(args, snap_dir)

            if args.start is None:
                args.start = find_last_snapshot(args, snap_dir)
            if ignore_start:
                args.end = args.start
            else:
                args.end = int(args.end)

            # prepare starformation file
            # star_form_filename = snap_restart_destination + '/star_form.ascii'
            # if args.verbose:
            #     print('====================================================')
            #     print('prepare star formation file "%s"...' % star_form_filename)
            #     sys.stdout.flush()
            # if os.path.exists(star_form_filename) and not args.overwrite:
            #     print('ERROR: file "%s" already ' % star_form_filename + \
            #                          'exists (consider "--overwrite")!', file=sys.stderr)
            #     sys.exit(1)
            # with open( star_form_filename, 'w' ) as f:
            #     print('%-12s \t %-11s \t %-18s \t %-10s \t %-11s' % (
            #                 'ID', 'a_form', 'r [kpc]', 'r/R200(a)', 'Z'), file=f)
            #
            # snap_num = args.start
            # print('prepare: snapshot ', snap_num, args.end)
            #
            # if args.verbose:
            #     print('found halo & galaxy of interest in %f sec' % (
            #             time.time() - start_time))

        # prepare trace folder after checking that input files are found
        if not os.path.exists(snap_restart_destination):
            if args.verbose: print('create profile folder: "%s"' % snap_restart_destination)
            os.makedirs(snap_restart_destination)
        # if os.listdir(snap_restart_destination):
        #     print('WARNING: profile is not empty!', file=sys.stderr)

        # do tracing / loop over snapshot (in reverse order)
        if args.verbose: start_time = time.time()

        print('*** process snapshots ', args.start, ' ... ', args.end)
        print('*** prepare')
        snap_exec = 'prepare'
        snap_num = args.start
        snap_start = args.start
        snap_end = args.end
        snap_overwrite = args.overwrite
        last_snap = None
        cmd_par = args.par
        cmd_par1 = args.par1
        cmd_par2 = args.par2
        cmd_par3 = args.par3
        cmd_par4 = args.par4
        cmd_par5 = args.par5
        cmd_destination = args.destination

        command_str = load_command(args.command)
        if command_str != '':
            print("*** execute command ", args.command)
            print("*** command parameter ", cmd_par)
            try:
                exec(command_str, globals(), locals())
            except Exception as e:
                print("*** error executing prepare ", e)


        if args.reverseorder:
            loop_start = snap_start
            loop_end = snap_end - 1
            loop_step = -args.step
        else:
            loop_start = snap_end
            loop_end = snap_start + 1
            loop_step = args.step

        for snap_num in range(loop_start, loop_end, loop_step):
            snap_fname = get_snap_fname(snap_dir, args.name_pattern, snap_num)
            filelist = glob.glob(snap_fname)
            if len(filelist) != 1:
                print('ERROR: pattern "%s" invalid ' % args.name_pattern,
                      'identifies more than one snapshot!', file=sys.stderr)
                sys.exit(1)
            else:
                snap_fname = filelist[0]
            if args.verbose:
                print('====================================================')
                print('process snapshot #%03d' % snap_num)
                print('process snapshot ', snap_fname, 'profile =', args.profile)
                sys.stdout.flush()

            if basedir != '' and snap_fname.find(basedir) == 0:
                findex = snap_fname.find(basedir)
                snap_fname_rel = snap_fname[len(basedir):]
                snap_cache = pg.SnapshotCache(snap_fname_rel, profile=args.profile)
            else:
                snap_fname_rel = ''
                snap_cache = pg.SnapshotCache(snap_fname, profile=args.profile)
            print('*** ', snap_num, ' process')
            # if args.findgxfast:
            #     snap_cache.load_snapshot(findgxfast=True)
            # else:
            if args.cachedonly:
                if not snap_cache.exists_in_cache():
                    print('*** skip snapshot ', snap_fname, '...profile =', args.profile)
                    continue
                print('*** use snapshot cache ', snap_fname, '...profile =', args.profile)
                load_OK = True
            else:
                print('*** load snapshot ', snap_fname, '...profile =', args.profile)
                try:
                    load_OK = True
                    if args.loadsf:
                        loadsf = True
                    else:
                        loadsf = False
                    snap_cache.load_snapshot(loadSF=loadsf)
                except Exception as e:
                    print("error loading snapshot ")
                    load_OK = False

            if load_OK:
                if args.withplot:
                    gx = snap_cache.galaxy
                    if gx is not None:
                        if not 'galaxy-all' in snap_cache.gx_properties:
                            if args.verbose:
                                print("plot particles in galaxy ", gx.parts)
                            buf = plot_snapshot(snap_cache, snap_cache.galaxy, 'Galaxy star particles')
                            snap_cache.gx_properties.append('galaxy-all', buf)
                        else:
                            if args.verbose:
                                print("plot exists already in cache")

                snap_exec = 'process'
                if command_str != '' and len(snap_cache.halo_properties) > 2: # no dummy halo
                    try:
                        exec(command_str, globals(), locals())
                    except Exception as e:
                        print("*** error executing process ", e)

                if args.updatecache:
                    print('*** ', snap_num, ' writing cache')
                    snap_cache.write_cache()
                    print('*** ', snap_num, ' finished')

                last_snap = snap_cache

                if args.verbose:
                    print('writing restart file "%s"...' % RESTART_FILENAME)
                    sys.stdout.flush()
                with open(RESTART_FILENAME, 'wb') as f:
                    pickle.dump((vars(args),snap_num), f)

                if args.verbose:
                    print('all done with snapshot #%03d in ' % snap_num + \
                          '%f sec' % (time.time() - start_time))
                    start_time = time.time()
                    sys.stdout.flush()

                if time.time()-t_start > pg.UnitScalar(args.tlimit, 's'):
                    if args.verbose:
                        print()
                        print('time limit (%s) exceeded -- stop!' % args.tlimit)
                    break

        print('*** close')
        snap_exec = 'close'
        if command_str != '':
            try:
                exec(command_str, globals(), locals())
            except Exception as e:
                print("*** error executing close ", e)

        if args.verbose:
            print()
            print('finished.')

else:
    print('unrecognized module name ', __name__ )



'''
Determine whether we are in interactive mode or not and provide function to
securely import h5py.
'''
__all__ = ['git_descr', 'interactive', 'module_dir', 'can_use_h5py',
           'secure_get_h5py', 'gc_full_collect']

import sys
import os
from .utils import *
import gc

module_dir = os.path.dirname(__file__)+'/'

def git_descr(path=os.curdir, dirty='dirty', PEP440=False):
    """
    Get a brief description of the current git revision of a directory.

    Args:
        path (str):     The path that point to or into the git repository you want
                        to get a revision / version description of.
        dirty (str):    If the repository is dirty, append '-<dirty>' to the
                        string. (I.e. call `git describe [...] --dirty=-<dirty>`.)
                        If None, do not mark a dirty repository.
        PEP440 (bool):  Convert to a PEP 0440 compliant version string plus a
                        leading 'v'. It is assumed that `git describe [...]`
                        returns a string that starts with a 'v' followed parts of
                        a PEP 0440 compliant version string. This typically
                        requires that the last git tag exists and is like 'v1.23'.

    Returns:
        descr (str):    A description of the current git revision of the repo.

    Raises:
        IOError:        If the given path does not exists.
        subprocess.CalledProcessError:
                        If the call of `git describe --always --tags [--dirty]`
                        failed.
        RuntimeError:   If PEP440 is True and the string obtained by `git
                        describe` returned could not have been converted into a
                        PEP 0440 compliant version.
    """
    if not os.path.exists(os.path.expanduser(path)):
        raise IOError('The path "%s" does not exist!' % path)
    if not os.path.isdir(path):
        path = os.path.dirname(path)
    import subprocess

    cmd = ['git', 'describe', '--always', '--tags']
    if dirty is not None:
        cmd.append('--dirty=-'+dirty)
    descr = subprocess.check_output(cmd, cwd=path).strip().decode('ascii')

    if PEP440:
        import re
        match = re.match(
                r'^(?P<vtag>(v|V)(\d+!)?\d+(\.\d+)*({a|b|rc}\d+)?(\.\d+)?)' + \
                r'(?P<rev>[a-zA-Z0-9\-]*)$',
                descr)
        if not match:
            raise RuntimeError('Description obtained by `git describe` is ' + \
                               'not PEP440 compliant. Probably git tags are ' + \
                               'missing or the last one is not a version ' +
                               'tag.')
        vtag = match.group('vtag')
        rev = match.group('rev')
        if re.match('-\d+-', rev):
            i = rev.replace('-','x',1).find('-')
            rev = rev[:i] + '+' + rev[i+1:]
        descr = vtag + rev.replace('-', '.')

    return descr

import __main__ as main
if hasattr(main, '__file__'):
    interactive = False
else:
    interactive = True

VERBOSE_QUIET       = 0
VERBOSE_TACITURN    = 1
VERBOSE_NORMAL      = 2
VERBOSE_TALKY       = 3
verbose = VERBOSE_NORMAL if interactive else VERBOSE_QUIET
allow_parallel_conversion = interactive

DEFAULT_Rvir_property = 'R200_com'      # Halo property to get virial radius
DEFAULT_gx_radius = 0.1                 # fraction of Rvir to get galaxy radius
DEFAULT_findgxfast = False              # default method to find zentral galaxy
DEFAULT_linking_length = "6 ckpc"           # linking length None for FoF from pepare_zoom, in samples '6 ckpc', gTrace 0.05
DEFAULT_linking_vel = '200 km/s'        # linking vel for FoF from prepare zoom, gtrace 100 km/s
DEFAULT_lowres_threshold = 1e-2         # theshold for FoF

# Plot parameters used by tool and utils
PLOT_figx = 11
PLOT_figy = 9
PLOT_fontsize = 18

class _h5py_dummy(object):
    '''
    Class holding some stuff to make pygad work even if there is no h5py
    module that can be loaded, but only this class to be returned by
    secure_get_h5py.
    '''
    class File(object):
        '''No object is of type h5py.File...'''
        def __init__(self, *args, **kwargs):
            pass
    @staticmethod
    def is_hdf5(filename):
        '''No file is detected as HDF5 file...'''
        return False

_can_use_h5py = False
def can_use_h5py():
    '''Test whether h5py was properly imported by secure_get_h5py.'''
    global _can_use_h5py
    return _can_use_h5py

@static_vars(_called=False)
def secure_get_h5py():
    '''
    Try to import h5py and return it. If this fails, return a dummy class, that
    has a class 'File' and a static function 'is_hdf5' in order to make code run,
    that, however, then can never really use h5py...
    '''
    global _can_use_h5py
    try:
        import h5py
        _can_use_h5py = True
        return h5py
    except:
        if not secure_get_h5py._called:
            print('WARNING: Could not import h5py -- format 3 ' + \
                                 'Gadget files are, hence, not supported.', file=sys.stderr)
            secure_get_h5py._called = True
        return _h5py_dummy

def gc_full_collect():
    '''Collect all collectable garbage.'''
    # just to be sure:
    old_thresholds = gc.get_threshold()
    gc.set_threshold(1, 1, 1)
    # sometimes a single call is not enough!
    while gc.collect():
        pass
    gc.set_threshold(*old_thresholds)
    if len(gc.garbage) > 0:
        print('WARNING: there is uncollectable garbage after a ' + \
                             'call of `gc_full_collect`!', file=sys.stderr)




'''
Module for general convenience routines for plotting.

Only very little doctests are possible, since it is mostly visual inspection
required...
    # 'plasma' and 'viridis' are not always available (they are from mpl v2.0 on)
    >>> for cmap in ['jet', 'rainbow', #'plasma', 'viridis',
    ...              'age', 'isolum']:
    ...     if isinstance(cmap, str):
    ...         cmap = plt.cm.get_cmap(cmap)
    ...     normed_cmap = isolum_cmap(cmap)
    ...     colors = normed_cmap(np.arange(normed_cmap.N))
    ...     lum = luminance(colors)
    ...     if np.any( np.abs(lum - np.mean(lum)) > 1e-6 ):
    ...         print('%s:'%cmap.name, np.mean(lum), np.percentile(lum,[0,100]))
'''
__all__ = ['CM_DEF', 'luminance', 'isolum_cmap', 'color_code', 'show_image',
           'scatter_map', 'make_scale_indicators', 'add_cbar']

import numpy as np
import matplotlib as mpl
import matplotlib.pyplot as plt
from matplotlib.colors import LinearSegmentedColormap, ListedColormap
from ..units import *
from ..binning import *
import sys

# 'jet' is bad -- the new default in matplotlib v2.0 ('viridis') is much better
CM_DEF = 'viridis'

# cf. http://alienryderflex.com/hsp.html
RGB_lum_weight = np.array([0.299, 0.587, 0.114])


def luminance(colors):
    '''
    Return the luminance of (a) color(s).

    Args:
        colors (array-like):    Either a single color in rgb space (shape (3,) or
                                (4,)) of an array of these (shape (N,3) of (N,4)).
                                The alpha channel is ignored.

    Returns:
        lum (np.ndarray):       The luminance of the color(s).
    '''
    colors = np.asarray(colors)
    if len(colors.shape) == 1:
        c = colors[:3]
    else:
        c = colors[:, :3]
    return np.sqrt(np.dot(c ** 2, RGB_lum_weight))


def isolum_cmap(cmap, isolum=1.0, desat=None):
    '''
    Return version of the colormap with constant luminance.

    Args:
        cmap (str, Colormap):   The colormap to norm in luminance.
        isolum (float):         The grade to which the colormap shall be converted
                                to a iso-luminosity one.
        desat (float):          If not None, a factor of how much to desatureate
                                the colormap first. This allows higher luminance.

    Returns:
        isolum (Colormap):      The iso-luminance colormap.
    '''
    if isinstance(cmap, str):
        cmap = plt.cm.get_cmap(cmap)
    colors = cmap(np.arange(cmap.N))

    if desat is not None:
        if not (0.0 <= desat <= 1.0):
            raise ValueError('`desat` needs to be in [0,1], but is %s!' % desat)
        hsv = mpl.colors.rgb_to_hsv(colors[:, :3])
        hsv[:, 1] *= 1.0 - desat
        colors[:, :3] = mpl.colors.hsv_to_rgb(hsv)

    # convert RGBA to perceived greyscale luminance
    # cf. http://alienryderflex.com/hsp.html
    if not (0.0 <= isolum <= 1.0):
        raise ValueError('`isolum` needs to be in [0,1], but is %s!' % isolum)
    lum = (isolum * luminance(colors)) + (1. - isolum)
    for i in range(3):
        colors[:, i] /= lum
    # if rgb=(1,1,1), then lum=sqrt(3)>1
    colors[:, :3] /= np.max(colors[:, :3])
    lum = luminance(colors)

    return mpl.colors.LinearSegmentedColormap.from_list(cmap.name + '_lumnormed',
                                                        colors, cmap.N)


def color_code(im_lum, im_col, cmap=CM_DEF, vlim=None, clim=None,
               zero_is_white=False):
    '''
    Colorcode a mono-chrome image with onother one.

    Args:
        im_lum (np.ndarray):    The mono-chrome image to color code. In other
                                words: the map of the brightness values. Needs to
                                be of shape (w,h).
        im_col (np.ndarray):    The image/map values that are converted into
                                colors. Needs to be of the same shape as
                                im_lum: (w,h).
        cmap (str, Colormap):   The colormap to use to convert the values of
                                im_col into actual colors.
        vlim (sequence):        The limits for brightnesses.
        clim (sequence):        The limits for the colors.
        zero_is_white (bool):   Instead of scaling the image luminance by im_lum,
                                desaturate the colors to white for im_lum values
                                approaching the lower limit of vlim (or zero if
                                this is None).

    Returns:
        im (np.ndarray):        The rgb image (shape (w,h,3)).
    '''
    if vlim is None:
        vlim = np.percentile(im_lum[np.isfinite(im_lum)], [0, 100])
    if clim is None:
        clim = np.percentile(im_col[np.isfinite(im_col)], [0, 100])

    im_lum = scale01(im_lum, vlim)
    im_col = scale01(im_col, clim)

    if isinstance(cmap, str): cmap = mpl.cm.get_cmap(cmap)
    im = cmap(im_col)

    if zero_is_white:
        white = np.ones(im.shape[:-1])
        im[:, :, 0] = im_lum * im[:, :, 0] + (1. - im_lum) * white
        im[:, :, 1] = im_lum * im[:, :, 1] + (1. - im_lum) * white
        im[:, :, 2] = im_lum * im[:, :, 2] + (1. - im_lum) * white
    else:
        im[:, :, 0] *= im_lum
        im[:, :, 1] *= im_lum
        im[:, :, 2] *= im_lum

    return im


def show_image(m, extent=None, cmap=CM_DEF, vlim=None, aspect=None,
               interpolation='nearest', ax=None, **kwargs):
    '''
    Show a map or an image with the 'physicsit's orientation'.

    The first axis of `m` is the x-direction and the second one the y-direction,
    contrary to the default behaviour of plt.imshow.

    Args:
        m (Map, array-like):    The map/image to show. It can be a luminance image
                                (shape (w,h)), a rgb-image (shape (w,h,3)) or a
                                rgba-image (shape (w,h,4)).
        extent (array-like):    The ranges of the image. It can either be a sequence
                                of four values (xmin,xmax,ymin,ymax) or an
                                array-like, of the structure
                                [[xmin,xmax],[ymin,ymax]].
        cmap (str, Colormap):
                            The colormap to use, it a luminance image was passed.
                            Ignored otherwise.
        vlim (sequence):    The minimum and maximum value of the image to show. If
                            you want to specify just one, pass [vmin,None] for
                            instance. Ignored if the image is not a luminance only
                            one (shape==(w,h)).
        aspect (str, float):The aspect ratio of the image (see ax.imshow for more
                            information).
        interpolation (str):The interpolation to use between pixels. See imshow
                            for more details.
        ax (AxesSubplot):   The axis object to plot on. If None, a new one is
                            created by plt.subplots().
        **kwargs:           Further arguments are passed to ax.imshow.

    Returns:
        fig (Figure):       The figure of the axis plotted on.
        ax (AxesSubplot):   The axis plotted on.
        im (AxesImage):     The image instance created.
    '''
    if len(m.shape) != 2 and not (len(m.shape) == 3 and m.shape[-1] in [3, 4]):
        raise ValueError('Image has to have shape (w,h), (w,h,3) or (w,h,4)!')
    if extent is None:
        extent = getattr(m, 'extent', None)
        if extent is None:
            raise ValueError('No extent given and `m` is not a `Map`!')

    if ax is None:
        fig, ax = plt.subplots()
    else:
        fig = ax.get_figure()
    vmin, vmax = vlim if vlim is not None else (None, None)
    if extent is not None:
        extent = np.asarray(extent).ravel()

    if len(m.shape) == 2:
        m = m.T
    else:
        m = np.dstack([m[:, :, i].T for i in range(m.shape[-1])])

    im = ax.imshow(m, origin='lower', extent=extent,
                   cmap=cmap, vmin=vmin, vmax=vmax,
                   aspect=aspect, interpolation=interpolation, **kwargs)

    return fig, ax, im


def scatter_map(x, y, s=None, qty=None, av=None, bins=150, extent=None,
                logscale=False, vlim=None, cmap=None, colors=None, colors_av=None,
                clim=None, clogscale=False, fontcolor=None, fontsize=14,
                outline=True, cbartitle=None, showcbar=True, aspect='auto',
                ax=None, zero_is_white=False, **kwargs):
    '''
    Do a binned scatter plot.

    Rather than doing a actual scatter plot, the data points are binned onto a
    grid (optionally color-coded by some quantity) and then the resulting map is
    displayed.

    Args:
        x (str, UnitArr):       The quantity to bin along the x-axis.
        y (str, UnitArr):       The quantity to bin along the y-axis.
        s (Snap):               The (sub-)snapshot from which to use the gas of.
                                Only needed, if one of x, y, qty, colors, or
                                colors_av is a string (block name).
        qty (str, array-like):  If given, not just scatter, but actually bin this
                                quantity onto a grid.
        av (str, array-like):   If given, plot the average the quantity `qty`,
                                weighted by this quantity.
        bins (int, sequence):   The number of bins per axis.
        extent (UnitArr):       The extent to plot over:
                                [[min(x),max(x)],[min(y),max(y)]]
        logscale (bool):        Whether to take the logarithm of the bins for
                                color-coding (or brightness, if colors are
                                provieded).
        vlim (sequence):        The limits for the bin values.
        cmap (str,Colormap):    The colormap to use.
        colors (str, UnitArr):  If passed, the bins are color-coded by this
                                quantity.
        colors_av (str, UnitArr):
                                If given, the colors quantity is averaged with
                                these weights.
        clim (sequence):        The limits for the color-coding.
        clogscale (bool):       Whether to do the color-coding logaritmically.
        fontcolor (str):        The color to use for the colorbar ticks and
                                labels.
        fontsize (int):         The size of the axis ticks and the cbar title.
                                Other font sizes are scaled accordingly.
        outline (array-like):   Draw an outline around text for better
                                readability. The first element is the thickness,
                                the second the color.
        cbartitle (str):        A specific colorbar title. If not given, a
                                automatic one is chosen.
        showcbar (bool):        Whether to show a colorbar for the color-coding.
        aspect (str, float):    The aspect ratio of the plot.
        ax (AxesSubplot):       The axis object to plot on. If None, a new one is
                                created by plt.subplots().
        zero_is_white (bool):   Instead of scaling the image luminance by `qty` if
                                if there is a color given, desaturate the colors
                                to white for `qty` values approaching the lower
                                limit of vlim (or zero if this is None).
        **kwargs:               Further arguments are passed to show_image.

    Returns:
        fig (Figure):       The figure of the axis plotted on.
        ax (AxesSubplot):   The axis plotted on.
        im (AxesImage):     The image instance created.
       [cbar (Colorbar):    The colorbar, if showcbar is True.]
    '''
    xname = ''
    yname = ''
    qtyname = ''
    cname = ''
    if isinstance(x, str):
        xname = x
        x = s.get(x)
    if isinstance(y, str):
        yname = y
        y = s.get(y)
    if isinstance(qty, str):
        qtyname = qty
        qty = s.get(qty)
    if isinstance(av, str):
        av = s.get(av)
    if cmap is None:
        cmap = CM_DEF if colors is None else 'isolum'
        cmap = mpl.cm.get_cmap(cmap)
        cmap.set_bad('w' if zero_is_white else 'k')
        if fontcolor is None:
            fontcolor = 'k' if zero_is_white else 'w'
    if isinstance(cmap, str):
        cmap = mpl.cm.get_cmap(cmap)
    if isinstance(colors, str):
        cname = colors
        colors = s.get(colors)
    if isinstance(colors_av, str):
        colors_av = s.get(colors_av)
    if fontcolor is None:
        fontcolor = 'k' if zero_is_white else 'w'

    if extent is None:
        extent = np.asarray([[np.min(x), np.max(x)], [np.min(y), np.max(y)]])
    else:
        if getattr(extent, 'units', None):
            extent = [extent[0], extent[1]]
            if getattr(x, 'units', None) is not None:
                extent[0] = extent[0].in_units_of(x.units)
            if getattr(y, 'units', None) is not None:
                extent[1] = extent[1].in_units_of(y.units)
        extent = np.asarray(extent)

    if av is not None:
        grid = gridbin2d(x, y, av * qty, bins=bins, extent=extent,
                         nanval=0.0)
        grid /= gridbin2d(x, y, av, bins=bins, extent=extent,
                          nanval=0.0)
        # grid[np.isnan(grid)] = 0.0
    else:
        grid = gridbin2d(x, y, qty, bins=bins, extent=extent, nanval=0.0)
    if getattr(vlim, 'units', None) is not None \
            and getattr(grid, 'units', None) is not None:
        vlim = vlim.in_units_of(grid.units)
    if logscale:
        grid = np.log10(grid)
        if vlim is not None:
            vlim = np.log10(vlim)
    if vlim is None:
        vlim = np.percentile(grid[np.isfinite(grid)], [1, 99])

    if colors is None:
        clim = vlim
    else:
        if colors_av is not None:
            col = gridbin2d(x, y, colors_av * colors, bins=bins, extent=extent,
                            nanval=0.0)
            col /= gridbin2d(x, y, colors_av, bins=bins, extent=extent,
                             nanval=0.0)
            col[np.isnan(col)] = 0.0
        else:
            col = gridbin2d(x, y, colors, bins=bins, extent=extent, nanval=0.0)

        if clogscale:
            col = np.log10(col)
            if clim is not None:
                clim = np.log10(clim)

        if clim is None:
            finitegrid = col[np.isfinite(col)]
            clim = [finitegrid.min(), finitegrid.max()]
            del finitegrid
        grid = color_code(grid, col, cmap=cmap, vlim=vlim, clim=clim,
                          zero_is_white=zero_is_white)

    fig, ax, im = show_image(grid, extent=extent, cmap=cmap, aspect=aspect, ax=ax,
                             clim=clim, **kwargs)

    if showcbar:
        if cbartitle is None:
            name = cname if colors is not None else qtyname
            count_units = r'$\mathrm{count}$'
            if colors is not None:
                units = getattr(colors, 'units', None)
            elif qty is not None:
                units = getattr(qty, 'units', None)
            else:
                units = count_units
            if units == 1 or units is None:
                units = ''
            elif units is not count_units:
                units = r'[$%s$]' % units.latex()
            cbartitle = name + (' ' if (name != '' and units != '') else '') + units
            if (logscale and colors is None) or \
                    (clogscale and colors is not None):
                cbartitle = r'$\log_{10}$(' + cbartitle + ')'

        cbar = add_cbar(ax, cbartitle, clim=clim, cmap=cmap, fontcolor=fontcolor,
                        fontsize=fontsize, fontoutline=outline, nticks=7)

    xunits = r'[$%s$]' % x.units.latex() if getattr(x, 'units', None) else ''
    yunits = r'[$%s$]' % y.units.latex() if getattr(y, 'units', None) else ''
    ax.set_xlabel('%s%s' % (xname, xunits), fontsize=fontsize)
    ax.set_ylabel('%s%s' % (yname, yunits), fontsize=fontsize)
    for tl in ax.get_xticklabels():
        tl.set_fontsize(0.8 * fontsize)
    for tl in ax.get_yticklabels():
        tl.set_fontsize(0.8 * fontsize)

    if showcbar:
        return fig, ax, im, cbar
    else:
        return fig, ax, im


def make_scale_indicators(ax, extent, scaleind='line', scaleunits=None,
                          xaxis=0, yaxis=1, fontsize=14, fontcolor='black',
                          outline=None):
    '''
    Set the scale indicators for a map.

    Args:
        ax (AxesSubplot):       The axis to set the indicators for.
        extent (UnitArr):       The extent plotted over. It needs to be of the
                                form: [[min(x),max(x)],[min(y),max(y)]]
        scaleind (str):         Can be:
                                    'labels':   ticks and labels are drawn on the
                                                axes
                                    'line':     a bar is drawn in the lower left
                                                corner indicating the scale
                                    'none':     neither nor is done -- make the
                                                axes unvisible
        scaleunits (str):       If the scale indication is 'labels', this is the
                                units of these. (Can still be None, though.)
        xaxis/yaxis (int):      The x- and y-axis in indices used to label the
                                axis in 'labels' scale mode.
        fontsize (int,float):   The font size to use.
        fontcolor (str):        The font color to use.
        outline (array-like):   Draw an outline around text for better
                                readability. The first element is the thickness,
                                the second the color.
    '''
    import matplotlib.patheffects

    if outline is True:
        outline = [
            3,
            np.array([1, 1, 1]) - mpl.colors.ColorConverter.to_rgb(fontcolor)
        ]
    if outline is None:
        path_effects = []
    else:
        path_effects = [mpl.patheffects.withStroke(linewidth=outline[0],
                                                   foreground=outline[1]),
                        mpl.patheffects.Normal()]
    if scaleind in [None, 'none']:
        ax.get_xaxis().set_visible(False)
        ax.get_yaxis().set_visible(False)
    elif scaleind == 'labels':
        if scaleunits is None:
            units_txt = ''
        else:
            units_txt = r' [$%s$]' % scaleunits.latex()
        x_label = r'$%s$%s' % ({0: 'x', 1: 'y', 2: 'z'}[xaxis], units_txt)
        ax.set_xlabel(x_label, fontsize=fontsize)
        y_label = r'$%s$%s' % ({0: 'x', 1: 'y', 2: 'z'}[yaxis], units_txt)
        ax.set_ylabel(y_label, fontsize=fontsize)
        for tl in ax.get_xticklabels():
            tl.set_fontsize(0.8 * fontsize)
        for tl in ax.get_yticklabels():
            tl.set_fontsize(0.8 * fontsize)
    elif scaleind == 'line':
        width = extent[:, 1] - extent[:, 0]
        scale = width.min() / 4.0
        order = 10.0 ** int(np.log10(scale))
        if scale / order < 2.0:
            scale = 1.0 * order
        elif scale / order < 5.0:
            scale = 2.0 * order
        else:
            scale = 5.0 * order
        scale_label = r'%g%s' % (scale,
                                 ' $%s$' % width.units.latex()
                                 if getattr(width, 'units', None) is not None
                                 else '')
        line = np.array([[extent[0, 0] + 0.05 * extent[0].ptp(),
                          extent[0, 0] + 0.05 * extent[0].ptp() + scale],
                         [extent[1, 0] + 0.12 * extent[1].ptp(),
                          extent[1, 0] + 0.12 * extent[1].ptp()]])
        if outline:
            ax.plot(line[0], line[1], color=outline[1], linewidth=3 + outline[0])
        ax.plot(line[0], line[1], color=fontcolor, linewidth=3)
        ax.text(np.mean(line[0]), extent[1, 0] + 0.10 * extent[1].ptp(),
                scale_label, color=fontcolor, size=0.9 * fontsize,
                horizontalalignment='center', verticalalignment='top',
                transform=ax.transData, path_effects=path_effects)
        ax.get_xaxis().set_visible(False)
        ax.get_yaxis().set_visible(False)
    else:
        import warnings
        warnings.warn('Unknown scaling indicator scaleind="%s"!' % scaleind)


def add_cbar(ax, cbartitle, clim, cmap=None, fontcolor='black', fontsize=14,
             fontoutline=None, nticks=None, tick_dist=None):
    '''
    Adding a pygad standard colorbar for a map.

    Args:
        ax (AxesSubplot):       The axis to set the indicators for.
        cbartitle (str):        The title for the color bar.
        clim (array-like):      The limits of the color bar.
        cmap (str, Colormap):   The colormap to use.
        fontcolor (str):        The font color for the ticks and the title.
        fontsize (int,float):   The font size of the title (ticks are 0.65 this
                                size).
        fontoutline (array-like):
                                Draw an outline around the text for better
                                readability. The first element is the thickness,
                                the second the color.
        nticks (int):           Set the number of ticks in the colorbar. If None,
                                it will be chosen automatically.
        tick_dist (float):      The distance between the tick labels, i.e. they
                                are placed at multiples of the given number.

    Returns:
        cbar (Colorbar):        The added colorbar instance.
    '''
    from mpl_toolkits.axes_grid1.inset_locator import inset_axes
    import matplotlib.patheffects

    if fontoutline is True:
        fontoutline = [
            3,
            np.array([1, 1, 1]) - mpl.colors.ColorConverter.to_rgb(fontcolor)
        ]
    if fontoutline is None:
        path_effects = []
    else:
        path_effects = [mpl.patheffects.withStroke(linewidth=fontoutline[0],
                                                   foreground=fontoutline[1]),
                        mpl.patheffects.Normal()]

    if isinstance(cmap, str):
        cmap = mpl.cm.get_cmap(cmap)
    cax = inset_axes(ax, width="70%", height="3%", loc=1)
    norm = mpl.colors.Normalize(vmin=clim[0], vmax=clim[1])

    cbar = mpl.colorbar.ColorbarBase(cax, cmap=cmap, norm=norm,
                                     orientation='horizontal')

    if tick_dist is not None:
        cbar.locator = mpl.ticker.MultipleLocator(tick_dist)
    if nticks is not None:
        cbar.locator = mpl.ticker.MaxNLocator(nbins=nticks)
    cbar.update_ticks()

    cbar.ax.tick_params(labelsize=8)
    for tl in cbar.ax.get_xticklabels():
        tl.set_path_effects(path_effects)
        tl.set_color(fontcolor)
        tl.set_fontsize(0.65 * fontsize)

    cbar.set_label(cbartitle, color=fontcolor, fontsize=fontsize, labelpad=12,
                   path_effects=path_effects)

    return cbar




'''
Module for convenience routines for plotting profiles.

Doctests impossible, since they would require visual inspection...
'''
__all__ = ['profile', 'history', 'SFR_history', 'flow_history']

import numpy as np
import matplotlib as mpl
import matplotlib.pyplot as plt
from ..units import *
from ..analysis import *
from .. import environment
from .. import utils
from .. import physics
import sys

def profile(s, Rmax, qty, av=None, units=None, dens=True, proj=None,
            N=50, logbin=False, minlog=None, logscale=True, ylabel=None,
            labelsize=14, ignoreZeros=False, ax=None, **kwargs):
    '''
    Plot a profile.

    Args:
        s (Snap):           The (sub-)snapshot to plot the quantity of.
        Rmax (UnitScalar):  The radius to plot out to.
        qty (str, UnitQty): The quantity to plot the profile for. It can be a
                            string that can be passed to s.get or a ready block
                            (of length len(s)).
        av (str, UnitQty):  The quantity to average 'qty' over. Otherwise as
                            'qty'.
        units (str, Unit):  The units in which the profile shall be plotted.
        dens (bool):        Whether to plot a density (surface density if proj is
                            not None, else volume density).
        proj (int):         Project along this axis and plot a cylindrical
                            profile. If None, however, plot a spherical profile.
                            (0:'x', 1:'y', 2:'z')
        N (int):            The number of bins.
        logbin (bool):      Whether to bin logarithmically in radius.
        minlog (UnitScalar):If logbin==True, this is the smallest bin edge (there
                            is one between 0 and this value, though).
                            Default: Rmax/100.
        logscale (bool):    Whether to plot y-values in log-scale.
        ylabel (str):       A custom y-axis label.
        labelsize (int):    The font size of the labels. The tick size will get
                            adjusted accordingly.
        ax (AxesSubplot):   The axis object to plot on. If None, a new one is
                            created by plt.subplots().
        ignoreZeros:        if set to True zero values are replaced by numpy.NaN
        **kwargs:           Further keyword arguments are passed to ax.plot (e.g.
                            'linewidth', 'color', or 'label').

    Returns:
        fig (Figure):       The figure of the axis plotted on.
        ax (AxesSubplot):   The axis plotted on.
    '''
    Rmax = UnitScalar(Rmax, s['pos'].units, subs=s)
    if logbin:
        if minlog is None:
            minlog = Rmax / 100.0
        else:
            minlog = UnitScalar(minlog, s['pos'].units)
        r_edges = np.logspace(np.log10(minlog), np.log10(float(Rmax)), N)
        r_edges = np.array( [0] + list(r_edges) )
        r = 10.0**((np.log10(r_edges[1:-1]) + np.log10(r_edges[2:])) / 2.0)
        r = np.array( [r_edges[1]/(r_edges[2]/r_edges[1])] + list(r) )
    else:
        r_edges = np.linspace(0, float(Rmax), N)
        r = (r_edges[:-1] + r_edges[1:]) / 2.0

    if dens:
        prof = profile_dens(s, qty, av, r_edges, proj=proj)
    else:
        prof = radially_binned(s, qty, av, r_edges, proj=proj)

    if units is not None:
        prof.convert_to(units, subs=s)

    if ax is None:
        fig, ax = plt.subplots()
    else:
        fig = ax.get_figure()

    if ignoreZeros:
        prof[prof==0] = np.NaN
    ax.plot(r, prof, **kwargs)

    if logbin:
        ax.set_xscale('log')
    if logscale:
        ax.set_yscale('log')

    ax.set_xlabel(r'$%s$ [$%s$]' % ('r' if proj is None else 'R',
                                    s['pos'].units.latex()),
                  fontsize=labelsize)
    if ylabel is None:
        name = ''
        if isinstance(av,str):
            name += str(av) + '-weighted '
        if dens:
            name = r'$\Sigma' if proj else r'$\rho'
            if isinstance(qty,str) and qty!='mass':
                name += r'_\mathrm{%s}$' % str(qty)
            else:
                name += r'$'
        else:
            name += str(qty) if isinstance(qty,str) else ''
        ylabel = r'%s [$%s$]' % (name, prof.units.latex())
    ax.set_ylabel(ylabel, fontsize=labelsize)

    for tl in ax.get_xticklabels():
        tl.set_fontsize(0.8*labelsize)
    for tl in ax.get_yticklabels():
        tl.set_fontsize(0.8*labelsize)

    return fig, ax

def history(s, qty, time=None, av=None, units=None, diff=False, N=50,
            t_edges=None, tlim=None, log=False, ylabel=None, labelsize=14,
            linewidth=2, add_z_ax=True, ax=None, **kwargs):
    '''
    Plot the archiological history of some stellar quantity.

    Args:
        s (Snap):               The (sub-)snapshot to plot the quantity of.
        qty (str, UnitQty):     The quantity to plot the profile for. It can be a
                                string that can be passed to s.get or a ready
                                block (of length len(s)).
        time (str, UnitQty):    The time to bin the quantity in. If None and the
                                (sub-)snapshot is stars only, it defaults to
                                'cosmic_time()-age'; otherwise it needs to be
                                given.
        av (str, UnitQty):      The quantity to average 'qty' over. Otherwise as
                                'qty'.
        units (str, Unit):      The units in which the profile shall be plotted.
        diff (boolean):         Whether to actually plot the time derivative.
        N (int):                The number of bins. It is ignored, if the edges
                                are given explicitly by `t_edges`.
        t_edges (UnitQty):      The time edges for the bins explicitly.
        tlim (list):            The limits in time in the plot. (Independent of
                                the bins used.)
        log (bool):             Plot in log-scale on the y-axis.
        ylabel (str):           A custom y-axis label.
        labelsize (int):        The font size of the labels. The tick size will
                                get adjusted accordingly.
        linewidth (int,float):  The linewidth.
        add_z_ax (boolean):     Whether the additional scale for redshifts is
                                added on top.
        ax (AxesSubplot):       The axis object to plot on. If None, a new one is
                                created by plt.subplots().
        **kwargs:               Further keyword arguments are passed to ax.plot
                                (e.g. 'linewidth', 'color', or 'label').

    Returns:
        fig (Figure):           The figure of the axis plotted on.
        ax (AxesSubplot):       The axis plotted on.
    '''
    now = s.cosmic_time()
    if isinstance(qty, str):
        Q = s.get(qty)
    else:
        Q = qty
    if len(s) != len(Q):
        raise ValueError('The length of the quantity array ' +
                         '(%s) ' % utils.nice_big_num_str(len(Q)) +
                         'does not match the length of the stellar ' +
                         '(sub-)snapshot (%s)!' % utils.nice_big_num_str(len(s)))
    if av is not None:
        if isinstance(av, str):
            AV = s.get(av)
        if len(s) != len(AV):
            raise ValueError('The length of the averaging quantity array ' +
                             '(%s) ' % utils.nice_big_num_str(len(AV)) +
                             'does not match the length of the stellar ' +
                             '(sub-)snapshot ' +
                             '(%s)!' % utils.nice_big_num_str(len(s)))
    if time is None:
        if len(s.stars) == len(s):
            time = 'cosmic_time()-age'
        else:
            raise ValueError('Time quantity is not defined!')
    if isinstance(time, str):
        time = s.get(time)
    if str(time.units).endswith('_form]'):
        from ..snapshot import age_from_form
        # only convert reasonable values & ensure not to overwrite blocks
        mask = (time!=-1) & np.isfinite(time)
        time = time.copy()
        new = now - age_from_form(time[mask],subs=s)
        time.units = new.units
        time[mask] = new

    if t_edges is None:
        t_edges = UnitArr(np.linspace(0,float(now),N+1), time.units)
        t_edges = UnitArr(np.linspace(0,float(now),N+1), time.units)
    else:
        t_edges = UnitQty(t_edges, time.units)

    Q_hist = []
    for t0, t1 in zip(t_edges[:-1], t_edges[1:]):
        mask = (t0<=time) & (time<t1)
        if av is None:
            Q_hist.append( Q[mask].sum() )
        else:
            av_hist.append( (Q[mask]*AV[mask]).sum() / AV[mask].sum() )
    Q_hist = UnitArr(Q_hist, Q.units)

    if diff:
        Q_hist /= t_edges[1:] - t_edges[:-1]

    if units is not None:
        Q_hist.convert_to(units, subs=s)

    if ax is None:
        fig, ax = plt.subplots()
    else:
        fig = ax.get_figure()

    t = (t_edges[:-1] + t_edges[1:]) / 2.0
    ax.plot(t, Q_hist, linewidth=linewidth, **kwargs)
    if tlim is None:
        tlim = UnitQty([t_edges.min(), t_edges.max()], units=t_edges.units)
    else:
        tlim = UnitQty(tlim, units=t_edges.units)
    ax.set_xlim(tlim)

    if add_z_ax:
        universe_age = s.cosmology.universe_age()
        z_minor = np.array([0, 0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9,
                            1, 1.5, 2, 3, 4, 5, 6, 7, 8, 9, 10])
        z_major = np.array([0, 0.5, 1, 2, 3, 4, 10])
        t_major = [universe_age-s.cosmology.lookback_time(z) for z in z_major]
        t_major = UnitArr(t_major, t_major[0].units).in_units_of(t.units)
        t_minor = [universe_age-s.cosmology.lookback_time(z) for z in z_minor]
        t_minor = UnitArr(t_minor, t_minor[0].units).in_units_of(t.units)
        ax_z = ax.twiny()
        ax_z.set_xlim(ax.get_xlim())
        ax_z.xaxis.set_tick_params(which='minor', size=4, width=1.5)
        ax_z.xaxis.set_tick_params(which='major', size=7, width=1.5)
        ax_z.set_xticks(t_minor, minor=True)
        ax_z.set_xticks(t_major)
        ax_z.set_xticklabels(['%g'%z for z in z_major])
        ax_z.set_xlabel(r'redshift $z$', fontsize=labelsize)
        for tl in ax_z.get_xticklabels():
            tl.set_fontsize(0.8*labelsize)
        for tl in ax_z.get_yticklabels():
            tl.set_fontsize(0.8*labelsize)
    ax.xaxis.set_tick_params(which='major', size=7, width=1.5)
    ax.xaxis.set_tick_params(which='minor', size=4, width=1.5)
    ax.yaxis.set_tick_params(which='major', size=7, width=1.5)
    ax.yaxis.set_tick_params(which='minor', size=3, width=1.5)

    if log:
        ax.set_yscale('log')

    ax.set_xlabel(r'cosmic time $t$ [$%s$]' % t.units.latex(),
                  fontsize=labelsize)
    if ylabel is None:
        name = ''
        if isinstance(av, str):
            name += str(av) + '-weighted '
        name += str(qty) if isinstance(qty,str) else ''
        ylabel = r'%s [$%s$]' % (name, Q_hist.units.latex())
    ax.set_ylabel(ylabel, fontsize=labelsize)

    for tl in ax.get_xticklabels():
        tl.set_fontsize(0.8*labelsize)
    for tl in ax.get_yticklabels():
        tl.set_fontsize(0.8*labelsize)

    return fig, ax

def SFR_history(s, units='Msol/yr', **kwargs):
    """
    Plot the archeological star formation history of the stars of a snapshot.

    This function basically calls `history` with qty='inim'.

    Args:
        s (Snap):               The (sub-)snapshot to plot the SFR history of.
        units (str, Unit):      The units in which the profile shall be plotted.
        **kwargs:               Further keyword arguments are passed to `history`.
                                Se its help for more.

    Returns:
        fig (Figure):           The figure of the axis plotted on.
        ax (AxesSubplot):       The axis plotted on.
    """
    s = s.stars
    units = Unit(units)
    kwargs['diff'] = kwargs.get('diff', True)
    kwargs['ylabel'] = kwargs.get('ylabel', r'$\dot M_*$ [$%s$]'%units.latex())
    fig, ax = history(s, 'inim',
                      units=units, **kwargs)
    return fig, ax

def flow_history(s, qty='mass', inout='infall', recycles='first/re',
                 z_max_min='3 kpc', t_out_min='100 Myr', max_cycle=np.inf,
                 units='Msol/yr', **kwargs):
    """
    Plot the in-/outflow rates as from the gas trace files.

    This function uses `history` in the background.

    Args:
        s (Snap):               The (sub-)snapshot to plot the SFR history of.
        qty ('mass'/'metals'):  Mass or metal flow rates.
        inout ('infall'/'ejection'):
                                Inflow rates ('infall') or outflow rates
                                ('ejection').
        recycles (str,int):     The options are:
                                * 'first/re':   plot the first in-/outflow and the
                                                re-accretion seperately together
                                                with the total in-/outflow rate
                                * 'all':        Plot the in-/outflow rates for
                                                each number of recycling.
                                * <int>:        Only plot the given number of
                                                first in-/outflows / recycles.
        z_max_min (UnitScalar): Only include in-/outflow if the cycle got higher
                                in z-direction than this limit.
        t_out_min (UnitScalar): Limit the included cycles by a minimum time
                                outside of the region of interest.
        max_cycle (int):        Limit the number of cycles considered.
        units (str, Unit):      The units in which the profile shall be plotted.
        ylabel (str):           A custom y-axis label.
        **kwargs:               Further keyword arguments are passed to `history`.
                                Se its help for more.

    Returns:
        fig (Figure):           The figure of the axis plotted on.
        ax (AxesSubplot):       The axis plotted on.
    """
    g = s.gas
    units = Unit(units)
    z_max_min = UnitScalar(z_max_min, 'kpc', subs=s)
    t_out_min = UnitScalar(t_out_min, 'Gyr', subs=s)
    kwargs['diff'] = kwargs.get('diff', True)
    kwargs['N'] = kwargs.get('N', 50)
    kwargs['ylabel'] = kwargs.get('ylabel', r'$\dot M$ [$%s$]'%units.latex())
    kwargs['linewidth'] = kwargs.get('linewidth', 2)

    if environment.verbose >= environment.VERBOSE_NORMAL:
        print('calculate the total %s rates...' % ('inflow' if inout=='infall'
                                                   else 'outflow'))
        sys.stdout.flush()
    max_N_cycle = min( np.max(g['num_recycled']), max_cycle )
    t_edges = np.linspace(0, float(g.cosmology.universe_age()), kwargs['N']+1)
    t_edges = UnitArr(t_edges, g.cosmology.universe_age().units)
    """
    mass = np.zeros(len(t_edges)-1, dtype=float)
    for i,t0,t1 in zip(np.arange(len(mass)),t_edges[:-1],t_edges[1:]):
        for n in range(max_N_cycle+1):
            tbin = (t0<=g[inout+'_time'][:,n]) & (g[inout+'_time'][:,n]<t1)
            mass[i] += g[qty+'_at_'+inout][tbin,n].sum()
    mass = UnitArr(mass, g[qty+'_at_'+inout].units)
    """

    t_lims = UnitArr([0,g.cosmology.universe_age()],
                     g.cosmology.universe_age().units)
    mass = UnitArr(np.zeros(kwargs['N'], dtype=float), g[qty+'_at_'+inout].units)
    from ..binning import gridbin
    for n in range(max_N_cycle+1):
        if n>1:
            sub = g[ (g['cycle_z_max'][:,n] > z_max_min) &
                     (g['out_time'][:,n-1] > t_out_min) ]
        else:
            sub = g[ (g['cycle_z_max'][:,n] > z_max_min) ]
        binned = gridbin(sub[inout+'_time'][:,n].reshape((len(sub),1)),
                         qty=sub[qty+'_at_'+inout][:,n],
                         bins=kwargs['N'],
                         extent=t_lims.reshape((1,2)))
        binned = binned.reshape((kwargs['N'],))
        if n==0:
            first_mass = binned
        mass += binned

    t = (t_edges[:-1]+t_edges[1:]) / 2.
    dt = t_edges[1:]-t_edges[:-1]

    if environment.verbose >= environment.VERBOSE_NORMAL:
        print('calculate the individual %s rates and plot...' % ('inflow'
                if inout=='infall' else 'outflow'))
        sys.stdout.flush()
    if 'ax' in kwargs:
        ax = kwargs['ax']
        fig = ax.get_figure()
    else:
        fig, ax = plt.subplots()
        kwargs['ax'] = ax

    pltargs = kwargs.copy()
    pltargs['linewidth'] = 1.4*kwargs['linewidth']
    for no in ['N', 'diff', 'ylabel', 'ax', 'add_z_ax', 'av']:
        if no in pltargs:
            del pltargs[no]
    ax.plot(t, (mass/dt).in_units_of(units,subs=s),
            label='total '+inout, **pltargs)

    if recycles == 'all':
        recycles = max_N_cycle
    if isinstance(recycles,int):
        for n in range(max_N_cycle+1):
            if n>1:
                sub = g[ (g['cycle_z_max'][:,n] > z_max_min) &
                         (g['out_time'][:,n-1] > t_out_min) ]
            else:
                sub = g[ (g['cycle_z_max'][:,n] > z_max_min) ]
            history(sub,
                    qty='%s_at_%s[:,%d]'%(qty,inout,n),
                    time='%s_time[:,%d]'%(inout,n),
                    units=units,
                    add_z_ax=(n==0),
                    label='%d. %s'%(n+1,inout), **kwargs)
    elif recycles == 'first/re':
        sub = g[ (g['cycle_z_max'][:,0] > z_max_min) ]
        history(sub,
                qty='%s_at_%s[:,0]'%(qty,inout),
                time='%s_time[:,0]'%(inout),
                units=units,
                add_z_ax=True,
                label='first %s'%inout, **kwargs)
        pltargs['linewidth'] = kwargs['linewidth']
        ax.plot(t, ((mass-first_mass)/dt).in_units_of(units,subs=s),
                label='re- '+inout, **pltargs)
    else:
        raise ValueError('`recycles` is %s -- not understood!' % recycles)

    ax.legend(loc='upper left')

    return fig, ax




'''
A UnitArr that belongs to a snapshot and can have derived arrays, which are
deleted from the snapshot, if the array changes.

Examples:
    >>> #from snapshot import Snapshot
    >>> from ..environment import module_dir
    >>> snap = Snapshot(module_dir+'snaps/snap_M1196_4x_470')
    >>> sa = SimArr([1,2,3], units='a kpc / h_0', dtype=float, snap=snap)
    >>> sa[:2]
    SimArr([ 1.,  2.], units="a kpc h_0**-1", snap="snap_M1196_4x_470")
    >>> sa.in_units_of('kpc')
    UnitArr([ 1.38888889,  2.77777778,  4.16666667], units="kpc")
    >>> sa.convert_to('kpc')
    >>> sa
    SimArr([ 1.38888889,  2.77777778,  4.16666667],
           units="kpc", snap="snap_M1196_4x_470")
    >>> sa *= snap.cosmology.h_0 / snap.scale_factor
    >>> assert sa.snap is snap

    # Operations with SimArr's, yield simple UnitArr's
    >>> 2*sa
    UnitArr([ 2.,  4.,  6.], units="kpc")
    >>> s = sa+sa
    >>> s
    UnitArr([ 2.,  4.,  6.], units="kpc")

    Results of operations with SimArr's shall be UnitArr and must not have any
    (hidden) references to the snapshot anymore!
    And similar shall hold for the return value of `in_units_of` (if actually a
    conversion happend and it is not just `self` returned).
    Both is done with the `downgrade_to_UnitArr` method.
    >>> sc = sa.in_units_of('AU', copy=True)
    >>> assert sc is not sa
    >>> sa.downgrade_to_UnitArr()
    >>> assert not isinstance(sa, SimArr)
    >>> for a in (sa, s, sc):
    ...     while a is not None:
    ...         assert not isinstance(a, SimArr)
    ...         assert not hasattr(a, '_snap')
    ...         assert not hasattr(a, '_dependencies')
    ...         a = a.base
'''

__all__ = ['SimArr']

import numpy as np
from ..units import *
from .snapshot import Snapshot
import functools
import weakref


class SimArr(UnitArr):
    '''
    A UnitArr that belongs to a snapshot and can have derived arrays, which are
    deleted from the snapshot, if the array changes.

    Operations on this array return UnitArr's, e.g. type(a*a) = units.UnitArr,
    even if a is a SimArr.
    '''

    def __new__(subtype, data, units=None, snap=None, subs=None, **kwargs):
        ua = UnitArr(data, units=units, subs=subs, **kwargs)

        new = ua.view(subtype)
        if snap is not None:
            new._snap = weakref.ref(snap)
        else:
            new._snap = getattr(data, '_snap', lambda: None)
        new._dependencies = getattr(data, '_dependencies', set())

        return new

    def __array_finalize__(self, obj):
        UnitArr.__array_finalize__(self, obj)
        self._snap = getattr(obj, '_snap', lambda: None)
        self._dependencies = getattr(obj, '_dependencies', set())

    def downgrade_to_UnitArr(self):
        '''
        Change this instance (inplace) into a UnitArr and remove all (also hidden)
        references to the snapshot (even though they are weakref's).
        '''
        a = self
        while a is not None:
            if isinstance(a, SimArr):  # this might be problematic
                a.__class__ = UnitArr  # to port to Python 3.x ?!
            if hasattr(a, '_snap'):         del a._snap
            if hasattr(a, '_dependencies'): del a._dependencies
            a = a.base

    def __array_wrap__(self, array, context=None):
        ua = UnitArr.__array_wrap__(self, array, context)
        # seems like if and only if context is None, its not from a ufunc but just
        # slicing/masking and/or setting inplace (i.e. from a __i*__ function like
        # __iadd__) -- in these cases, we don't want references to the snapshot
        if context is not None:
            # might even already be a UnitArr, but the reference can still exist!
            ua.view(SimArr).downgrade_to_UnitArr()
        return ua

    @property
    def snap(self):
        '''The associated snapshot, if any, otherwise None.'''
        return self._snap()

    @property
    def dependencies(self):
        '''A set of the names of the derived blocks that depend on this one.'''
        return self._dependencies

    def __repr__(self):
        r = super(SimArr, self).__repr__()
        r = r.replace('UnitArr', 'SimArr').replace('\n ', '\n')
        if self.snap is not None:
            r = r[:-1] + (', snap=%s)' % self.snap.descriptor)
            if len(r) - r.rfind('\n') > 80:
                right = r.find('dtype=')
                if right == -1:
                    right = r.find('units=')
                    if right == -1:
                        right = r.find('snap=')
                arr_end = r.rfind('],') + 2
                if arr_end == 1: arr_end = r.find(',') + 1
                r = r[:arr_end] + '\n' + ' ' * 7 + r[right:]
        return r

    def __copy__(self, *a):
        if a:
            duplicate = UnitArr.__copy__(self, *a).view(SimArr)
        else:
            duplicate = UnitArr.__copy__(self).view(SimArr)
        duplicate._snap = self._snap
        duplicate._dependencies = self._dependencies.copy()
        return duplicate

    def __deepcopy__(self, *a):
        duplicate = UnitArr.__deepcopy__(self).view(SimArr)
        duplicate._snap = self._snap
        duplicate._dependencies = self._dependencies.copy()
        return duplicate

    def convert_to(self, units, subs=None):
        '''See UnitArr for documentation.'''
        if subs is None:
            subs = self.snap
        # super(SimArr, self).convert_to(units, subs=subs)
        UnitArr.convert_to(self, units, subs=subs)

    def in_units_of(self, units, subs=None, copy=False):
        '''See UnitArr for documentation. This, however, returns a UnitArr view.'''
        if subs is None:
            subs = self.snap
        conv = UnitArr.in_units_of(self, units, subs=subs, copy=copy)
        if conv is not self:
            # did an actual conversion with new array, that shall no longer be a
            # SimArr and shall not reference back the snapshot:
            conv.downgrade_to_UnitArr()
        return conv

    def invalidate_dependencies(self):
        '''
        Invalidate all dependencies by simply deleting them from the snapshot.
        (They get rederived automatically, if needed, anyway.)

        There is no need to call this function directly. It gets called once the
        array changes.
        '''
        for dep in self._dependencies:
            host = self.snap.get_host_subsnap(dep)
            if dep in host._blocks:
                del host[dep]


# __i*__ functions shall not erase the _snap and _dependencies attributes and
# shall invalidate all dependent SimArr's
def _ichange_wrapper(f):
    def ichange__(self, other):
        self.invalidate_dependencies()
        return f(self, other)

    ichange__.__name__ = f.__name__
    return ichange__


for fn in ('__iadd__', '__isub__', '__imul__', '__idiv__',
           '__itruediv__', '__ifloordiv__', '__imod__',
           '__ipow__', '__ilshift__', '__irshift__',
           '__iand__', '__ior__', '__ixor__'):
    setattr(SimArr, fn, _ichange_wrapper(getattr(UnitArr, fn)))


# __set*__ functions shall invalidate all dependant SimArr's
def _set_wrapper(f):
    def set__(self, *y, **kw):
        self.invalidate_dependencies()
        if kw:
            f(self, *y, **kw)
        else:
            f(self, *y)

    set__.__name__ = f.__name__
    return set__


for fn in ('__setitem__', '__setslice__'):
    setattr(SimArr, fn, _set_wrapper(getattr(UnitArr, fn)))




# these packages are loaded by gcache3 by default
import os
import io
import sys
import re
import time
import numpy as np
import matplotlib.pyplot as plt
import pygad as pg

# additional packages should be loaded in prepare section
# here your code of functions ONLY starts
def split_cols(col_names):

    col_list = col_names.split(',')
    return col_list

def write_header(col_list, fcat):
    print("CmdCatalog - write headline")
    line = "filename, Snapshot"
    for col in col_list:
        line += "," + col
    line += "\n"
    fcat.write(line)

def write_line(snapfile, col_list, fcat):
    line = snap_fname + "," + str(snap_num)
    for col in col_list:
        if col in snapfile.gx_properties:
            value = snapfile.gx_properties[col]
            if isinstance(value, pg.UnitArr):
                value = float(value.view(np.ndarray))
        else:
            value = " "
        line += "," + str(value)
    line += "\n"
    fcat.write(line)
    return


##############################################################################
# here start the gcache3 sections, invoking your code                        #
##############################################################################
# the order of the gcache3 sections is : init, prepare, process, close, end  #
# sections are identified by line starting with # gcache3-<section name>:    #
# the order must not be modified                                             #
##############################################################################

##############################################################################
# gcache3-init: ignored by gcache3
##############################################################################
pg.environment.verbose=pg.environment.VERBOSE_QUIET
snap_cache = pg.SnapshotCache("/mnt/hgfs/Astro/Magneticum/M0094/AGN/snap_m0094_sf_x_2x_094",
                          profile='FastRun')
snap_cache.load_snapshot()

snap_fname = "/mnt/hgfs/Astro/Magneticum/M0094/AGN/snap_m0094_sf_x_2x_094"
snap_num = 94
snap_start = 94
snap_end = 94
snap_overwrite = True
cmd_destination = os.getenv('SNAPCACHE_HOME', None)
cmd_par = "catalog.txt"
cmd_par1 = 'a,redshift,R200,M200,M_stars,R_half_M,R_half_M(faceon),R_eff(faceon),D/T(stars),jzjc_baryons,gamma-halo'

##############################################################################
# gcache3-prepare: processes before first snapshot
##############################################################################
#print("command CmdCatalog - prepare ", snap_num, " von ", snap_start, " bis ", snap_end)
pass

##############################################################################
# gcache3-process: processes for each snapshot
##############################################################################

col_names = split_cols(cmd_par1)
if cmd_par == '':
    fname = "catalog.txt"
else:
    fname = cmd_par

if cmd_destination is None:
    fname = snap_cache.get_profile_path() + fname
else:
    fname = cmd_destination + fname

if os.path.exists(fname):
    neu = False
else:
    neu = True

if neu:
    fcat = open(fname,"w")
    write_header(col_names, fcat)
else:
    fcat = open(fname, "a+")

#print("command CmdCatalog - write snapshot ", snap_num, " to catalog ", fname)
write_line(snap_cache, col_names, fcat)
fcat.close()

##############################################################################
# gcache3-close: processed after last snapshot
##############################################################################

#print("command CmdCatalog - finished ")
pass

##############################################################################
# gcache3-end: ignored by gcache3
##############################################################################





'''
In this module there is class defined that decorates a numpy array with units.

Note:
    Since the unit is global to the array, the speed penality is not too big.
    However, if one accesses elements, slices, etc. of a UnitArr, it might help to
    work with 'a.view(np.ndarray)' and add the units again at the end, rather than
    directly working with the UnitArr 'a'.

Examples:
    >>> v = UnitArr([[1,-2], [-3,1], [2,3]], dtype=np.float64, units='m/s')
    >>> v
    UnitArr([[ 1., -2.],
             [-3.,  1.],
             [ 2.,  3.]], units="m s**-1")
    >>> v.units
    Unit("m s**-1")
    >>> v[0]
    UnitArr([ 1., -2.], units="m s**-1")
    >>> v[0,0]
    1.0

    >>> v = UnitArr([1.5,-2.,3.], units='m/s')
    >>> w = UnitArr(v, 'km/h')
    >>> w
    UnitArr([  5.4,  -7.2,  10.8], units="km h**-1")
    >>> v.in_units_of('km/h')   # does not change v
    UnitArr([  5.4,  -7.2,  10.8], units="km h**-1")
    >>> v
    UnitArr([ 1.5, -2. ,  3. ], units="m s**-1")
    >>> v.convert_to('km/h')    # convertion in-place
    >>> v
    UnitArr([  5.4,  -7.2,  10.8], units="km h**-1")
    >>> UnitArr(UnitArr(v, 'm/s'), 'm/s')
    UnitArr([ 1.5, -2. ,  3. ], units="m s**-1")
    >>> v
    UnitArr([  5.4,  -7.2,  10.8], units="km h**-1")
    >>> v[:2].convert_to('m/s') # for consistency: convert entire array!
    >>> v
    UnitArr([ 1.5, -2. ,  3. ], units="m s**-1")
    >>> v.units = 'kg'  # just setting the units, no conversion
    >>> v
    UnitArr([ 1.5, -2. ,  3. ], units="kg")
    >>> sub = v[1:]
    >>> v[:2].units = 's'   # for consistency: convert entire array!
    >>> v
    UnitArr([ 1.5, -2. ,  3. ], units="s")
    >>> sub     # units are changed for the view as well!
    UnitArr([-2.,  3.], units="s")

    >>> v + v
    UnitArr([ 3., -4.,  6.], units="s")
    >>> v + UnitArr('1 min')
    UnitArr([ 61.5,  58. ,  63. ], units="s")
    >>> v *= 2.3
    >>> v
    UnitArr([ 3.45, -4.6 ,  6.9 ], units="s")
    >>> v -= [1.05, -5.1, 2.9]     # understood as in current units
    >>> v
    UnitArr([ 2.4,  0.5,  4. ], units="s")
    >>> v * v
    UnitArr([  5.76,   0.25,  16.  ], units="s**2")
    >>> v
    UnitArr([ 2.4,  0.5,  4. ], units="s")
    >>> v * np.array([1,2,3])
    UnitArr([  2.4,   1. ,  12. ], units="s")
    >>> np.array([1,2,3]) * v
    UnitArr([  2.4,   1. ,  12. ], units="s")
    >>> np.sqrt(v)
    UnitArr([ 1.54919334,  0.70710678,  2.        ], units="s**1/2")
    >>> np.round(v**Fraction(1,3), 3)   # no floats allowed, however, Fraction are!
    array([1.339,  0.794,  1.587])
    >>> np.mean(v)
    UnitArr(2.3, units="s")
    >>> v.sum()
    UnitArr(6.9, units="s")
    >>> v.prod()
    UnitArr(4.8, units="s**3")
    >>> for prop in [np.sum, np.cumsum, np.mean, np.std, np.median, np.ptp,
    ...              np.transpose, np.min, np.max]:
    ...     res = prop(v)
    ...     assert res.units == v.units
    ...     assert res._unit_carrier is not v._unit_carrier
    ...     b = res.base
    ...     while hasattr(b, '_unit_carrier'):
    ...         assert b._unit_carrier is res._unit_carrier
    ...         b = b.base
    ...     while b is not None:
    ...         assert not hasattr(b, '_unit_carrier')
    ...         b = b.base
    >>> M_earth = UnitArr('5.972e24 kg')
    >>> R_earth = UnitArr('12740 km', dtype=float) / 2.0
    >>> v_earth = UnitArr('29.78 km/s')
    >>> M_earth / (4/3.*np.pi * R_earth**3)
    UnitArr(5.515856e+12, units="kg km**-3")
    >>> (0.5 * M_earth * v_earth**2).in_units_of('J')
    UnitArr(2.648129e+33, units="J")
    >>> UnitArr('1.2 N*m').in_base_units()
    UnitArr(1.200000e+03, units="g m**2 s**-2")

    Unit conversion at construction:
    >>> UnitArr('10**8 km', units='AU')
    UnitArr(0.668458712227, units="AU")
    >>> UnitArr(UnitArr([1e8, 1e9], 'km'), units='AU')
    UnitArr([ 0.66845871,  6.68458712], units="AU")

    conversion to native types
    >>> s = UnitScalar(1.2, 'm')
    >>> float(s)
    1.2
    >>> int(s)
    1
    >>> l = UnitArr([2.4], 'km')
    >>> float(l)
    2.4
    >>> s / l
    UnitArr([ 0.5], units="km**-1 m")
    >>> float(s/l)
    0.0005
    >>> float( UnitArr([1.2], 'km/cm') )
    120000.0

    Pickling:
    >>> import pickle
    >>> v_pickled = pickle.dumps(UnitArr([1,2,3], 'inch'))
    >>> v_unpickled = pickle.loads(v_pickled)
    >>> v_unpickled.units
    Unit("inch")

    Some tests for underlying implementation
    >>> v = UnitArr([1.5,-2.,3.], units='m/s')
    >>> assert v.copy()._unit_carrier is not v._unit_carrier
    >>> assert v.astype(float)._unit_carrier is not v._unit_carrier
    >>> assert v.T._unit_carrier is not v._unit_carrier
    >>> assert (v+v)._unit_carrier is not v._unit_carrier
'''
__all__ = ['UnitArr', 'UnitQty', 'UnitScalar']

import numpy as np
import numpy.core.umath_tests
from . import units
from .units import *
from .units import _UnitClass
from fractions import Fraction
import warnings
import functools
import numbers


def UnitQty(obj, units=None, subs=None, dtype=None, copy=False):
    '''
    Convert to a UnitArr with enshured units.

    Args:
        obj (UnitArr, array-like, Unit, str, float):
                                The object to convert.
        units (Units, str):     If obj had units, convert to these, otherwise set
                                these units.
        subs (dict, Snap):      Substitutions as passe to UnitArr.convert_to.
        dtype (np.dtype, str):  The requested dtype.

    Returns:
        obj (UnitArr):          A unit array with the desired properties.

    Examples:
        >>> a = UnitArr([1,2,3], 'm')
        >>> UnitQty(a, units='m') is a
        True
        >>> UnitQty(a, copy=True) is a
        False
        >>> UnitQty(a, units='cm', dtype=float)
        UnitArr([ 100.,  200.,  300.], units="cm")
        >>> a
        UnitArr([1, 2, 3], units="m")
        >>> a = UnitArr([1.,2.,3.], 'm')
        >>> UnitQty(a, units='cm')
        UnitArr([ 100.,  200.,  300.], units="cm")
        >>> a
        UnitArr([ 1.,  2.,  3.], units="m")
        >>> UnitQty([1,2,3.5], units='s')
        UnitArr([ 1. ,  2. ,  3.5], units="s")
        >>> UnitQty('2.5 AU')
        UnitArr(2.5, units="AU")
        >>> UnitQty('0.5 min', units='s')
        UnitArr(30.0, units="s")
        >>> UnitQty('2.5 ckpc/h_0', 'kpc', subs=dict(a=0.35, h_0=0.7))
        UnitArr(1.25, units="kpc")
        >>> UnitQty(Unit('2.5 AU'))
        UnitArr(2.5, units="AU")
        >>> UnitQty([2.5, -0.3])
        UnitArr([ 2.5, -0.3])
        >>> UnitQty(2.5, 'AU')
        UnitArr(2.5, units="AU")
        >>> UnitQty('2.5')
        UnitArr(2.5)
        >>> UnitQty('2.5', 'min')
        UnitArr(2.5, units="min")
        >>> UnitQty({}) # doctest:+ELLIPSIS
        Traceback (most recent call last):
        ...
        ValueError: Only native dtype allowed for UnitArr, not object!
    '''
    if isinstance(obj, UnitArr) and not copy:
        if (units is None or obj.units == units) and \
                (dtype is None or obj.dtype == dtype):
            return obj
    return UnitArr(obj, units=units, subs=subs, dtype=dtype, copy=True)


def UnitScalar(obj, units=None, subs=None, dtype=None):
    '''
    Convert to a UnitArr with enshured units of a single number.

    Calls UnitQty and tests for the shape of the object.

    Args:
        obj (UnitArr, Unit, str, float):
                                The object to convert.
        units (Units, str):     If obj had units, convert to these, otherwise set
                                these units.
        subs (dict, Snap):      Substitutions as passe to UnitArr.convert_to.
        dtype (np.dtype, str):  The requested dtype.

    Raises:
        ValueError:             If the obj is actually an array.

    Examples:
        >>> UnitScalar(1.2, 'm')
        UnitArr(1.2, units="m")
        >>> UnitScalar('2.5 ckpc/h_0', 'kpc', subs=dict(a=0.35, h_0=0.7))
        UnitArr(1.25, units="kpc")
        >>> UnitScalar([1.2], 'm')
        Traceback (most recent call last):
        ...
        ValueError: object is not a scalar!
        >>> UnitScalar([1,2,3])
        Traceback (most recent call last):
        ...
        ValueError: object is not a scalar!
    '''
    obj = UnitQty(obj, units=units, subs=subs, dtype=dtype)
    if obj.shape:
        raise ValueError('object is not a scalar!')
    return obj


def with_own_units(func):
    '''Decorate a function return value with the units of the first argument.'''

    def u_func(self, *args, **kwargs):
        a = UnitQty(func(self.view(np.ndarray), *args, **kwargs))
        a._set_units_and_carrier_on_base(self.units)
        return a

    u_func.__name__ = func.__name__
    u_func.__doc__ = func.__doc__
    return u_func


class UnitArr(np.ndarray):
    '''
    A numpy array decorated with units.

    Args:
        data (array-like):  Either 'raw data' passed to the underlying numpy
                            array or a UnitArr. In the latter case, units are
                            taken from it.
        units (Unit, str):  The units of the array. If not None, they overwrite
                            those from data.
        keyword arguments:  Passed on to the array factory function of numpy
                            (np.array).

    New attributes:
        units (Unit):       The units of the data (can be None!).
    '''

    _ufunc_registry = {}

    def __new__(subtype, data, units=None, subs=None, **kwargs):
        # handle cases where `obj` is a string of a unit:
        if isinstance(data, str):
            data = Unit(data)
        if isinstance(data, _UnitClass):
            if units is None:
                data, units = data.scale, data.free_of_factors()
            else:
                if len(data._composition) == 0:
                    data = float(data)
                else:
                    data = data.in_units_of(units, subs=subs)

        # actually create the new object
        if isinstance(data, UnitArr) and 'copy' not in kwargs:
            # avoid interference of different `_unit_carrier`s and units
            kwargs['copy'] = True
        obj = np.array(data, **kwargs).view(subtype)
        # bool needed in some situations (such as np.median(UnitArr))
        if obj.dtype.kind not in 'uifb':
            raise ValueError('Only native dtype allowed for UnitArr, not %s!' %
                             obj.dtype)

        # all UnitArr's need to have a _unit_carrier
        obj._unit_carrier = obj
        obj._units = getattr(data, 'units', None)

        # bring data into desired units
        if units is not None:
            if obj.units is not None:
                obj.convert_to(units, subs=subs)
            else:
                obj.units = units

        return obj

    def __array_finalize__(self, obj):
        self._unit_carrier = getattr(obj, '_unit_carrier', self)
        self._units = getattr(obj, 'units', None)

    def __array_wrap__(self, array, context=None):
        if context is None:
            return array

        try:
            ufunc = context[0]
            units = UnitArr._ufunc_registry[ufunc](*context[1])
            if units is not None:
                units = units.gather()
                units._composition.sort()
            array = array.view(UnitArr)
            array._set_units_and_carrier_on_base(units)
            return array
        except KeyError:
            if not ufunc.__name__.endswith('(vectorized)'):
                warnings.warn('Operation \'%s\' on units is ' % ufunc.__name__ + \
                              '*not* defined! Return normal numpy array.')
            return np.asarray(array)
        except:
            raise

    @property
    def units(self):
        '''The units of the array.'''
        return self._units

    @units.setter
    def units(self, value):
        if value is not None:
            value = Unit(value)
        self._units = value

    def _set_units_and_carrier_on_base(self, units=None):
        uc = self
        if isinstance(self.base, UnitArr):
            uc = self.base._set_units_and_carrier_on_base(units)
            self._units = uc._units
        else:
            self._units = None if units is None else Unit(units)
        self._unit_carrier = uc

        return uc

    def __float__(self):
        try:
            a = self.astype(float).in_units_of(1)
        except:
            a = self
        return float(a.view(np.ndarray))

    def __int__(self):
        try:
            a = self.astype(float).in_units_of(1)
        except:
            a = self
        return int(a.view(np.ndarray))

    def __copy__(self, *a):
        if a:
            dupl = np.ndarray.copy(self, *a)
        else:
            dupl = np.ndarray.copy(self)
        dupl._set_units_and_carrier_on_base(self.units)
        return dupl

    def __deepcopy__(self, *a):
        dupl = np.ndarray.__deepcopy__(self).view(UnitArr)
        dupl._set_units_and_carrier_on_base(self.units)
        return dupl

    def copy(self, order=None):
        '''Create a real copy.'''
        if order:
            return self.__copy__(order)
        else:
            return self.__copy__()

    def __repr__(self, val=None):
        if not self.shape and self.dtype.kind == 'f':
            r = 'UnitArr('
            if val is None:
                f = float(self.view(np.ndarray))  # avoid conversion of units!
                r += str(f) if (1e-3 <= f <= 1e3) else ('%e' % f)
            else:
                r += str(val)
            if self.dtype not in ['int', 'float']:
                r += ', dtype=' + str(self.dtype)
            r += ')'
        else:
            if val is None:
                r = repr(self.view(np.ndarray))
                r = r[r.find('('):].replace('\n', '\n  ')
            else:
                r = '(' + str(val) + ')'
            r = 'UnitArr' + r
        if hasattr(self, 'units'):
            if self.units is not None and self.units != 1:
                r = r[:-1] + (', units="%s")' % str(self.units)[1:-1])
                if len(r) - r.rfind('\n') > 82:
                    right = r.find('dtype=')
                    if right == -1:
                        right = r.find('units=')
                    arr_end = r.rfind('],') + 2
                    if arr_end == 1: arr_end = r.find(',') + 1
                    r = r[:arr_end] + '\n' + ' ' * 8 + r[right:]
        else:
            r = r[:-1] + ', no units!)'
        return r

    def __str__(self):
        if not self.shape and self.dtype.kind == 'f':
            f = float(self.view(np.ndarray))  # avoid conversion of units!
            s = str(f) if (1e-3 <= f <= 1e3) else ('%e' % f)
        else:
            s = str(self.view(np.ndarray))
        if self.units is not None and self.units != 1:
            s += ' %s' % self.units
        return s

    # needed for pickling
    def __reduce__(self):
        pickled_state = np.ndarray.__reduce__(self)
        return (pickled_state[0], pickled_state[1],
                pickled_state[2] + (self.units,))

    # needed for unpickling
    def __setstate__(self, state):
        self.units = state[-1]
        np.ndarray.__setstate__(self, state[:-1])

    def astype(self, dtype, *args, **kwargs):
        '''
        Convert the type of the array.

        Internally np.ndarray.astype is called. See the documentation for more
        details.
        '''
        new = np.ndarray.astype(self, dtype, *args, **kwargs)
        if new is not self:
            new._set_units_and_carrier_on_base(self.units)
        return new

    def convert_to(self, units, subs=None):
        '''
        Convert the array into other units in place.

        If this is just a view of a bigger underlying UnitArr, this entire
        underlying is converted in order to avoid different units within the same
        array.

        Args:
            units (Unit, str):  The units to convert to.
            subs (Snap, dict):  Substitutions to be made during conversion. (For
                                more information see _UnitClass.substitute). If
                                this is a Snap, the values for the redshift 'z',
                                the scale factor 'a', and the Hubble parameter
                                'h_0' from it are used as well as the cosmology,
                                if not set explicitly by the other argument.

        Raises:
            UnitError:          In case the current units and the target units are
                                not convertable.
        '''
        units = Unit(units)
        if self.units == units:
            if str(self.units) != str(units):
                self.units = units
            return
        elif self.units is None:
            self.units = units
            return

        from ..snapshot.snapshot import Snapshot
        if isinstance(subs, Snapshot):
            snap, subs = subs, {}
            subs['a'] = snap.scale_factor
            subs['z'] = snap.redshift
            subs['h_0'] = snap.cosmology.h_0

        uc = self._unit_carrier
        if uc.dtype.kind != 'f':
            raise RuntimeError('Cannot convert UnitArr inplace, that do not ' + \
                               'have floating point data type!')
        fac = self._units.in_units_of(units, subs=subs)
        view = uc.view(np.ndarray)
        view *= fac
        self._units = units

    def in_units_of(self, units, subs=None, copy=False):
        '''
        Return the array in other units.

        Args:
            units (Unit, str):  The target units.
            subs (Snap, dict):  See 'convert_to'.
            copy (bool):        If set to False, return the array itself, if the
                                target units are the same as the current ones.

        Returns:
            converted (UnitArr):    The new, converted array. If copy is False and
                                    the target units are the same as the current
                                    ones, however, simply self is returned.

        Raises:
            UnitError:          In case the current units and the target units are
                                not convertable.
        '''
        if not isinstance(units, (_UnitClass, str, numbers.Number)):
            raise TypeError('Cannot convert type %s to units!' % \
                            type(units).__name__)
        units = Unit(units)
        if not copy and units == self.units:
            self.units = units
            return self
        c = self.copy()
        c.convert_to(units=units, subs=subs)
        return c

    def convert_to_base_units(self, subs=None):
        '''
        Convert to the base units, i.e. the units all other units are defined in.

        For more information see `convert_to`.
        '''
        units = self.units.standardize().free_of_factors()
        return self.convert_to(units, subs=subs)

    def in_base_units(self, subs=None, copy=False):
        '''
        Return the array in the base units, i.e. the units all other units are
        defined in.

        For more information see `in_units_of`.
        '''
        units = self.units.standardize().free_of_factors()
        return self.in_units_of(units, subs=subs, copy=copy)

    def __setitem__(self, i, value):
        # if both value and self have units and they are different,
        # take care of them
        if isinstance(value, UnitArr) and value.units is not None \
                and self.units is not None and value.units != self.units:
            np.ndarray.__setitem__(self, i, value.in_units_of(self.units))
        elif isinstance(value, _UnitClass) and self.units is not None and \
                value != self.units:
            np.ndarray.__setitem__(self, i, value.in_units_of(self.units))
        else:
            np.ndarray.__setitem__(self, i, value)

    def __setslice__(self, a, b, value):
        self.__setitem__(slice(a, b), value)

    @staticmethod
    def ufunc_rule(for_ufunc):
        def x(fn):
            UnitArr._ufunc_registry[for_ufunc] = fn
            return fn

        return x

    def _generic_add(self, x, op):
        return op(self, UnitQty(x, self.units))

    def __add__(self, x):
        return self._generic_add(x, op=np.add)

    def __sub__(self, x):
        return self._generic_add(x, op=np.subtract)

    def __iadd__(self, x):
        return self._generic_add(x, op=np.ndarray.__iadd__)

    def __isub__(self, x):
        return self._generic_add(x, op=np.ndarray.__isub__)

    def __mul__(self, x):
        return np.multiply(self, UnitQty(x))

    def __div__(self, x):
        return np.divide(self, UnitQty(x))

    def __truediv__(self, x):
        return np.true_divide(self, UnitQty(x))

    def __floordiv__(self, x):
        return np.floor_divide(self, UnitQty(x))

    def __mod__(self, x):
        return np.remainder(self, UnitQty(x, self.units))

    def __imul__(self, x):
        x = UnitQty(x)
        if self.units is None:
            self.units = x.units
        elif x.units is not None:
            self.units *= x.units
        np.ndarray.__imul__(self, x)
        return self

    def __idiv__(self, x):
        x = UnitQty(x)
        if x.units is not None:
            if self.units is None:
                self.units = 1 / x.units
            else:
                self.units /= x.units
        np.ndarray.__idiv__(self, x)
        return self

    def __itruediv__(self, x):
        x = UnitQty(x)
        if x.units is not None:
            if self.units is None:
                self.units = 1 / x.units
            else:
                self.units /= x.units
        np.ndarray.__itruediv__(self, x)
        return self

    def __ifloordiv__(self, x):
        x = UnitQty(x)
        if x.units is not None:
            if self.units is None:
                self.units = 1 / x.units
            else:
                self.units /= x.units
        np.ndarray.__ifloordiv__(self, x)
        return self

    def __imod__(self, x):
        x = UnitQty(x, self.units)
        np.ndarray.__imod__(self, x)
        return self

    def __pow__(self, x):
        if isinstance(x, Fraction) and self.dtype.kind == 'f':
            # way faster to use floats:
            res = np.ndarray.__pow__(self.view(np.ndarray), float(x))
            res = UnitQty(res, units=self.units ** x)
        else:
            res = np.ndarray.__pow__(self, x)
        return res

    def __ipow__(self, x):
        if isinstance(x, Fraction) and self.dtype.kind == 'f':
            # way faster to use floats:
            np.ndarray.__ipow__(self.view(np.ndarray), float(x))
        else:
            np.ndarray.__ipow__(self, x)
        self.units **= x
        return self

    def prod(self, axis=None, *args, **kwargs):
        x = np.ndarray.prod(self, axis, *args, **kwargs).view(UnitArr)
        if self.units is not None:
            if axis is None:
                x.units = self.units ** np.prod(self.shape)
            else:
                x.units = self.units ** self.shape[axis]
        return x

    def cumprod(self, axis=None, *args, **kwargs):
        raise NotImplementedError('Units of a cumulative product would not ' + \
                                  'be constant!')

    def var(self, *args, **kwargs):
        x = np.ndarray.var(self, *args, **kwargs).view(UnitArr)
        if self.units is not None:
            x.units = self.units ** 2
        return x

    @with_own_units
    def sum(self, *args, **kwargs):
        return np.ndarray.sum(self, *args, **kwargs)

    @with_own_units
    def cumsum(self, *args, **kwargs):
        return np.ndarray.cumsum(self, *args, **kwargs)

    @property
    def T(self):
        return self.transpose()

    @with_own_units
    def transpose(self, *args, **kwargs):
        return np.ndarray.transpose(self, *args, **kwargs)

    @with_own_units
    def byteswap(self, *args, **kwargs):
        return np.ndarray.byteswap(self, *args, **kwargs)

    @with_own_units
    def conj(self, *args, **kwargs):
        return np.conj(self, *args, **kwargs)

    @with_own_units
    def conjugate(self, *args, **kwargs):
        return np.conjugate(self, *args, **kwargs)

    @with_own_units
    def min(self, *args, **kwargs):
        return np.ndarray.min(self, *args, **kwargs)

    @with_own_units
    def max(self, *args, **kwargs):
        return np.ndarray.max(self, *args, **kwargs)

    @with_own_units
    def mean(self, *args, **kwargs):
        return np.mean(self, *args, **kwargs)

    @with_own_units
    def std(self, *args, **kwargs):
        return np.std(self, *args, **kwargs)

    @with_own_units
    def ptp(self, *args, **kwargs):
        return np.ptp(self, *args, **kwargs)

    def argpartition(self, *args, **kwargs):
        return np.ndarray.argpartition(self.view(np.ndarray), *args, **kwargs)

    def argsort(self, *args, **kwargs):
        return np.ndarray.argsort(self.view(np.ndarray), *args, **kwargs)


for f in (np.ndarray.__lt__, np.ndarray.__le__, np.ndarray.__eq__,
          np.ndarray.__ne__, np.ndarray.__gt__, np.ndarray.__ge__):
    # N.B. cannot use functools.partial because it doesn't implement the
    # descriptor protocol
    @functools.wraps(f, assigned=("__name__", "__doc__"))
    def wrapper_function(self, other, comparison_op=f):
        try:
            if hasattr(self, 'snap'):  # e.g. for SimArr
                other = UnitQty(other, self.units, subs=self.snap)
            else:
                other = UnitQty(other, self.units)
        except units.UnitError as e:
            if not e.msg.endswith('convertable!'):
                return NotImplemented
            other = UnitQty(other)
            if isinstance(other, UnitArr):
                raise units.UnitError('%r and %r ' % (self.units, other.units) +
                                      'are not comparible.')
        except:
            raise
        res = comparison_op(self, other)
        return res


    setattr(UnitArr, f.__name__, wrapper_function)


@UnitArr.ufunc_rule(np.add)
@UnitArr.ufunc_rule(np.subtract)
@UnitArr.ufunc_rule(np.minimum)
@UnitArr.ufunc_rule(np.maximum)
def _same_units_binary(a, b):
    a_units = getattr(a, 'units', None)
    b_units = getattr(b, 'units', None)
    if a_units is not None and b_units is not None:
        if a_units == b_units:
            return a_units
        else:
            a_units.in_units_of(b_units)
            raise units.UnitError('Units are not the same but they are ' +
                                  'convertable. Use operator instead of ' +
                                  'function or convert manually.')
    else:
        return a_units if a_units is not None else b_units


@UnitArr.ufunc_rule(np.negative)
@UnitArr.ufunc_rule(np.abs)
@UnitArr.ufunc_rule(np.floor)
@UnitArr.ufunc_rule(np.ceil)
@UnitArr.ufunc_rule(np.trunc)
@UnitArr.ufunc_rule(np.round)  # TODO: does not work, since the
@UnitArr.ufunc_rule(np.around)  # function does something more inbetween
@UnitArr.ufunc_rule(np.round_)  # than others (and calls np.rint)
@UnitArr.ufunc_rule(np.rint)
@UnitArr.ufunc_rule(np.fix)
@UnitArr.ufunc_rule(np.transpose)
@UnitArr.ufunc_rule(np.conjugate)
def _same_units_unary(a):
    return a.units


@UnitArr.ufunc_rule(np.multiply)
@UnitArr.ufunc_rule(np.cross)
@UnitArr.ufunc_rule(np.dot)
@UnitArr.ufunc_rule(np.core.umath_tests.inner1d)
def _mul_units(a, b):
    a_units = getattr(a, 'units', None)
    b_units = getattr(b, 'units', None)
    if a_units is not None and b_units is not None:
        ab_units = a_units * b_units
        return ab_units
    elif a_units is not None:
        return a_units
    else:
        return b_units


@UnitArr.ufunc_rule(np.divide)
@UnitArr.ufunc_rule(np.true_divide)
@UnitArr.ufunc_rule(np.floor_divide)
def _div_units(a, b):
    a_units = getattr(a, 'units', None)
    b_units = getattr(b, 'units', None)
    if a_units is not None and b_units is not None:
        ab_units = a_units / b_units
        return ab_units
    elif a_units is not None:
        return a_units
    else:
        return 1 / b_units


@UnitArr.ufunc_rule(np.reciprocal)
def _inv_units(a):
    if a.units is not None:
        return a.units ** -1
    else:
        return None


@UnitArr.ufunc_rule(np.remainder)
def _remainder_units(a, b):
    a_units = getattr(a, 'units', None)
    b_units = getattr(b, 'units', None)
    if a_units is not None and b_units is not None and a_units != b_units:
        a_units.in_units_of(b_units)
        raise units.UnitError('Units are not the same but they are ' +
                              'convertable. Use operator instead of ' +
                              'function or convert manually.')
    return a_units


@UnitArr.ufunc_rule(np.sqrt)
def _sqrt_units(a):
    if a.units is not None:
        return a.units ** Fraction(1, 2)
    else:
        return None


@UnitArr.ufunc_rule(np.square)
def _square_units(a):
    if a.units is not None:
        return a.units ** 2
    else:
        return None


@UnitArr.ufunc_rule(np.greater)
@UnitArr.ufunc_rule(np.greater_equal)
@UnitArr.ufunc_rule(np.less)
@UnitArr.ufunc_rule(np.less_equal)
@UnitArr.ufunc_rule(np.equal)
@UnitArr.ufunc_rule(np.not_equal)
@UnitArr.ufunc_rule(np.logical_and)
@UnitArr.ufunc_rule(np.logical_or)
@UnitArr.ufunc_rule(np.logical_xor)
@UnitArr.ufunc_rule(np.logical_not)
def _comparison_units(a, b):
    return None


@UnitArr.ufunc_rule(np.bitwise_and)
@UnitArr.ufunc_rule(np.bitwise_or)
@UnitArr.ufunc_rule(np.bitwise_xor)
@UnitArr.ufunc_rule(np.invert)
def _bitwise_op_units(a, b):
    return None


@UnitArr.ufunc_rule(np.power)
def _pow_units(a, b):
    if getattr(a, 'units', None) is not None:
        return a.units ** b
    else:
        return None


@UnitArr.ufunc_rule(np.arctan)
@UnitArr.ufunc_rule(np.arctan2)
@UnitArr.ufunc_rule(np.arcsin)
@UnitArr.ufunc_rule(np.arccos)
@UnitArr.ufunc_rule(np.arcsinh)
@UnitArr.ufunc_rule(np.arccosh)
@UnitArr.ufunc_rule(np.arctanh)
@UnitArr.ufunc_rule(np.sin)
@UnitArr.ufunc_rule(np.tan)
@UnitArr.ufunc_rule(np.cos)
@UnitArr.ufunc_rule(np.sinh)
@UnitArr.ufunc_rule(np.tanh)
@UnitArr.ufunc_rule(np.cosh)
def _trig_units(*a):
    return None


@UnitArr.ufunc_rule(np.exp)
@UnitArr.ufunc_rule(np.log)
@UnitArr.ufunc_rule(np.log2)
@UnitArr.ufunc_rule(np.log10)
def _exp_log_units(*a):
    return None


@UnitArr.ufunc_rule(np.sign)
def _no_units(*a):
    return None


@UnitArr.ufunc_rule(np.isinf)
@UnitArr.ufunc_rule(np.isneginf)
@UnitArr.ufunc_rule(np.isposinf)
@UnitArr.ufunc_rule(np.isnan)
@UnitArr.ufunc_rule(np.isfinite)
def _testfunc_units(*a):
    return None




'''
Module snapshots transformations.

Also doctest other parts of this sub-module:
    >>> import doctest
    >>> doctest.testmod(transformation)
    TestResults(failed=0, attempted=54)
'''
from .transformation import *




'''
Module for handling Gadget files.

Also doctest other parts of this sub-module:
    >>> import doctest
    >>> doctest.testmod(config)
    TestResults(failed=0, attempted=10)
    >>> doctest.testmod(lowlevel_file)
    TestResults(failed=0, attempted=12)

    handler module is implicitly tested in the snapshot module...
'''
from .config import *
from . import lowlevel_file
from .handler import *
import os

from ..environment import module_dir
read_config(['./gadget.cfg',
             os.getenv("HOME")+'/.config/pygad/gadget.cfg',
             module_dir+'config/gadget.cfg'])




'''
Calculating the cooling rates for gas particles using the Wiersma cooling table as
can be obtained from their website (http://www.strw.leidenuniv.nl/WSS08/).
The website is down, we have a copy on our bitbucket (https://bitbucket.org/broett/pygad/downloads/).

Doctests:
    >>> from ..environment import module_dir
    >>> from ..snapshot import Snapshot
    >>> from ..tools import prepare_zoom
    >>> s,h,g = prepare_zoom(module_dir+'snaps/snap_M1196_4x_470',
    ...                      star_form=None,
    ...                      gas_trace=None) # doctest: +ELLIPSIS
    prepare zoomed-in <Snap "snap_M1196_4x_470"; N=2,079,055; z=0.000>
    read info file from: .../snaps/trace/info_470.txt
    ...
    R200: 207.996... [kpc]
    M200: 1.08...e+12 [Msol]
    orientate at the galactic red. inertia tensor from info file
    ...
    M*:   2.93...e+10 [Msol]

    Get the cooling tables of not present:
    >>> import os
    >>> if not os.path.exists(module_dir+'CoolingTables/z_0.000.hdf5'):
    ...     url = 'https://bitbucket.org/broett/pygad/downloads/'
    ...     file = 'z_0.000_highres.tar.gz'
    ...     subprocess.run('wget -q  %s%s' % (url, file), check=True, shell=True)
    ...     subprocess.run('tar zxvf %s -C %s/../' % (file, module_dir), check=True, shell=True)
    ...     subprocess.run('rm -f %s' % file, check=True, shell=True)

    Calculate the cooling rates for just every 100th particle for reasons of
    runtime.
    >>> tbl = Wiersma_CoolingTable(module_dir+'CoolingTables/z_0.000.hdf5') # doctest: +ELLIPSIS
    reading cooling tables from "...CoolingTables/z_0.000.hdf5"
    >>> Lambda = tbl.get_cooling(s.gas[::100],
    ...                 units='erg cm**3 s**-1') # doctest: +ELLIPSIS
    calculate cooling rates for
      <Snap "snap_M1196_4x_470":gas[::100]; N=9,218; z=0.000>
    ...
    interpolate cooling rates for...
      Carbon...
    derive block C... done.
    ...
    interpolate ne/nH and solar metallicity ne/nH...
    combine to netto cooling rates...
    add the Compton cooling rates...
    >>> assert Lambda.units == 'erg cm**3 s**-1'

    The percentiles are still close to the overall ones, though
    >>> perc = np.percentile(Lambda, [10,25,50,75,90])
    >>> np.round(perc, decimals=25)
    array([-1.31e-23, -6.40e-24, -7.00e-25,  1.57e-23,  4.30e-23])

    The cooling rates of the galaxy are much smaller than they are overall:
    >>> Lambda_gal = tbl.get_cooling(g.gas,
    ...                 units='erg cm**3 s**-1') # doctest: +ELLIPSIS
    calculate cooling rates for
      <Snap "snap_M1196_4x_470":Ball(... z=0.000>
    interpolate cooling rates for...
      Carbon...
      ...
      Iron...
      H+He...
    interpolate ne/nH and solar metallicity ne/nH...
    combine to netto cooling rates...
    add the Compton cooling rates...
    >>> perc_gal = np.percentile(Lambda_gal, [10,25,50,75,90])
    >>> np.round(np.array(perc_gal)/np.array(perc), decimals=5)
    array([ 0.03983,  0.0583 ,  0.32214, -0.00594,  0.00067])
'''
__all__ = ['Wiersma_CoolingTable']

import numpy as np
from ..units import UnitQty, UnitArr
from .. import environment

h5py = environment.secure_get_h5py()

FULL_ELEMENT_NAME = {
    'H': 'Hydrogen',
    'He': 'Helium',
    'C': 'Carbon',
    'N': 'Nitrogen',
    'O': 'Oxygen',
    'Ne': 'Neon',
    'Mg': 'Magnesium',
    'Al': 'Aluminium',
    'Si': 'Silicon',
    'P': 'Phosphorus',
    'S': 'Sulphur',
    'Ca': 'Calcium',
    'Fe': 'Iron',
}
SHORT_ELEMENT_NAME = {v: k for k, v in FULL_ELEMENT_NAME.items()}
# US and Canada do not know English...
SHORT_ELEMENT_NAME['Aluminum'] = 'Al'


class Wiersma_CoolingTable(object):
    '''
    Handle HDF5 cooling tables from Wiersma et al. (2008) as taken from their
    website: http://www.strw.leidenuniv.nl/WSS08/
    The link is down but the same data are available on our bitbucket: https://bitbucket.org/broett/pygad/downloads/

    Args:
        path (str):     The path to the HDF5 file to load the table from.
    '''

    def __init__(self, path):
        self._path = path
        self._read()

    def _read(self, verbose=None):
        '''
        Read the tables and additional information from the HDF5 file.

        Args:
            verbose (int):          The verbosity level. Defaults to
                                    `environment.verbose`.
        '''
        if verbose is None:
            verbose = environment.verbose
        if verbose >= environment.VERBOSE_NORMAL:
            print('reading cooling tables from "%s"' % self._path)
        with h5py.File(self._path, 'r') as f:
            self._note = f['Header/Note'][()][0].decode('ascii')
            self._reference = f['Header/Reference'][()][0].decode('ascii')
            self._version = f['Header/Version'][()][0].decode('ascii')
            if verbose >= environment.VERBOSE_TALKY:
                print('  note:       "%s"' % self._note)
                print('  reference:  "%s"' % self._reference)
                print('  version:    "%s"' % self._version)

            self._redshift = f['Header/Redshift'][()][0]

            spec = f['Header/Species_names'][()]
            self._species = spec.astype('|U10')
            if verbose >= environment.VERBOSE_TALKY:
                print('  %2d species: %s' % (len(self._species),
                                             ', '.join(self._species)))

            if verbose >= environment.VERBOSE_TALKY:
                print('  reading cooling tables for')
                print('    metal free gas & electron density...')
            tbls = {
                'noZ': f['Metal_free/Net_Cooling'][()],
                'ne_nH': f['Metal_free/Electron_density_over_n_h'][()],
            }
            f['Metal_free/Mean_particle_mass'][()]
            for el in self._species:
                if verbose >= environment.VERBOSE_TALKY:
                    print('    %s...' % el)
                tbls[el] = f['%s/Net_Cooling' % el][()]
            if verbose >= environment.VERBOSE_TALKY:
                print('  reading table bins')
            self._nH_bins = f['Metal_free/Hydrogen_density_bins'][()]
            self._T_bins = f['Metal_free/Temperature_bins'][()]
            self._fHe_bins = f['Metal_free/Helium_mass_fraction_bins'][()]
            if verbose >= environment.VERBOSE_TALKY:
                print('    nH [cm^-3]: ', self.nH_range)
                print('    T [K]:      ', self.T_range)
                print('    fHe:        ', self.fHe_range)

            if verbose >= environment.VERBOSE_TALKY:
                print('  reading solar abundancies')
            sspec = f['Header/Abundances/Abund_names'][()]
            self._solar_species = sspec.astype('|U10')
            self._solar_mass_frac = \
                f['Header/Abundances/Solar_mass_fractions'][()]
            self._solar_ne_nH = f['Solar/Electron_density_over_n_h'][()]

        self._tbls = tbls

        def interp(x, y, z, kind='linear'):
            from scipy.interpolate import interp2d
            return interp2d(x, y, z.T, kind=kind, copy=True, bounds_error=True)

        self._metals_interp = {
            metal: interp(self._T_bins, self._nH_bins, self._tbls[metal])
            for metal in self._species
        }
        self._noZ_interp = [
            interp(self._T_bins, self._nH_bins, self._tbls['noZ'][ifHe])
            for ifHe in range(len(self._fHe_bins))
        ]
        self._ne_nH_interp = [
            interp(self._T_bins, self._nH_bins, self._tbls['ne_nH'][ifHe])
            for ifHe in range(len(self._fHe_bins))
        ]
        self._solar_ne_nH_interp = \
            interp(self._T_bins, self._nH_bins, self._solar_ne_nH)

    @property
    def note(self):
        return self._note

    @property
    def version(self):
        return self._version

    @property
    def reference(self):
        return self._reference

    @property
    def redshift(self):
        return self._redshift

    @property
    def species(self):
        return tuple(self._species)

    @property
    def nH_range(self):
        return self._nH_bins[0], self._nH_bins[-1]

    @property
    def T_range(self):
        return self._T_bins[0], self._T_bins[-1]

    @property
    def fHe_range(self):
        return self._fHe_bins[0], self._fHe_bins[-1]

    def _interpolate_fHe(self, interpolations, T, nH, fHe):
        '''
        For either `self._noZ_interp` or `self._ne_nH_interp`.
        '''
        assert interpolations is self._noZ_interp or \
               interpolations is self._ne_nH_interp
        from numbers import Number
        if isinstance(fHe, Number) or fHe.shape == tuple() or len(fHe) == 1:
            fHe = fHe * np.ones(len(T), dtype=float)
        fHe = np.array(fHe).astype(float)
        assert T.shape == nH.shape == fHe.shape

        ifHe = np.searchsorted(self._fHe_bins, fHe) - 1
        if np.any(fHe < self._fHe_bins[ifHe]) or \
                np.any(ifHe >= len(self._fHe_bins) - 1):
            raise IndexError('m_He/m_H (fHe) out of range')
        iTnH = list(zip(ifHe, T, nH))
        Qs_1 = np.array([
            interpolations[i](t, n) for i, t, n in iTnH
        ]).reshape(len(T))
        iTnH = list(zip(ifHe + 1, T, nH))
        Qs_2 = np.array([
            interpolations[i](t, n) for i, t, n in iTnH
        ]).reshape(len(T))
        alpha = (fHe - self._fHe_bins[ifHe]) \
                / (self._fHe_bins[ifHe + 1] - self._fHe_bins[ifHe])
        return (1. - alpha) * Qs_1 + alpha * Qs_2

    def get_cooling_for_species(self, species, T, nH, fHe=0.25):
        '''
        Get the cooling rate(s) of a given species / element in units of
        [erg cm^3 s^-1]: Lambda / n_H^2

        Note:
            These cooling rates do take the heating from the UV/X-ray background
            into account!

        Args:
            species (str):      The element to get the cooling rate(s) for. It can
                                either be one of the species listed in
                                `self.species` or 'noZ'/'metal-free'/'H+He'/'HHe'
                                for the cooling of Hydrogen and Helium only.
            T (UnitQty):        The temperature of the gas.
            nH (UnitQty):       The Hydrogen number density
            fHe (float):        The mass fraction of Helium. Only used if it is
                                asked for the cooling rate(s) for Hydrogen +
                                Helium.

        Returns:
            Lambda (UnitArr):   The cooling rate(s) for the given species in units
                                of [erg cm**3 s^-1].
        '''
        T = UnitQty(T, 'K').view(np.ndarray)
        nH = UnitQty(nH, 'cm**-3').view(np.ndarray)

        if species in FULL_ELEMENT_NAME:
            species = FULL_ELEMENT_NAME[species]
        if species in SHORT_ELEMENT_NAME:
            if len(T) == len(nH) and len(T) > 1:
                Lambda = np.array([
                    self._metals_interp[species](t, n) for t, n in zip(T, nH)
                ]).reshape(len(T))
            else:
                Lambda = self._metals_interp[species](T, nH)
        elif species in ['noZ', 'metal-free', 'H+He', 'HHe']:
            # additional interpolation for HHe is needed
            Lambda = self._interpolate_fHe(self._noZ_interp, T, nH, fHe)
        else:
            raise ValueError("Unknown species '%s'" % species)

        return UnitArr(Lambda, 'erg cm**3 s**-1')

    def get_cooling(self, s, Compton_cooling=True, units='erg cm**3 s**-1',
                    verbose=None):
        '''
        Calculate the cooling rates for all gas elements in a given
        (sub-)snapshot: Lambda / n_H^2

        Note:
            These cooling rates do take the heating from the UV/X-ray background
            into account!

        Args:
            s (Snap):               The snapshot for whichs gas to calculate the
                                    cooling rates.
            Compton_cooling (bool): Whether to also include the Compton cooling
                                    from the cosmic microwave background (CMB).
            units (Unit, str):      The units in which the result shall be
                                    returned.
            verbose (int):          The verbosity level. Defaults to
                                    `environment.verbose`.

        Returns:
            Lambda (UnitArr):       The cooling rates (Lambda/n_H**2) for each gas
                                    particle in the (sub-)snapshot `s`.
        '''
        if verbose is None:
            verbose = environment.verbose
        if abs(s.redshift - self.redshift) > 0.01:
            raise RuntimeError("The snapshot's redshift (z=%0.2f) " % s.redshift +
                               "does not match the one of the tables " +
                               "(z=%0.2f)!" % self._redshift)
        g = s.gas if (len(s.gas) != len(s) or len(s) == 0) else s
        if verbose >= environment.VERBOSE_NORMAL:
            print('calculate cooling rates for\n  %s' % g)

        T = g['temp'].in_units_of('K', subs=s).view(np.ndarray)
        nH = g['nH'].in_units_of('cm**-3', subs=s).view(np.ndarray)
        fHe = (g['He'] / (g['He'] + g['H'])).in_units_of(1, subs=s).view(np.ndarray)

        T[T < self.T_range[0]] = self.T_range[0]
        T[T > self.T_range[1]] = self.T_range[1]
        nH[nH < self.nH_range[0]] = self.nH_range[0]
        nH[nH > self.nH_range[1]] = self.nH_range[1]
        fHe[fHe < self.fHe_range[0]] = self.fHe_range[0]
        fHe[fHe > self.fHe_range[1]] = self.fHe_range[1]

        if verbose >= environment.VERBOSE_NORMAL:
            print('interpolate cooling rates for...')
        metal_cool = np.zeros(len(g), dtype=float)
        for species in self._species:
            try:
                if verbose >= environment.VERBOSE_NORMAL:
                    print('  %s...' % species)
                el = SHORT_ELEMENT_NAME[species]
                solar_abund = \
                    self._solar_mass_frac[np.where(self._solar_species == species)][0]
                el_cool = self.get_cooling_for_species(el, T, nH, fHe) \
                          * (g[el] / g['mass']) / solar_abund
                metal_cool += el_cool
            except ValueError as e:
                print("%s: %s" % (type(e).__name__, e.message))
        if verbose >= environment.VERBOSE_NORMAL:
            print('  H+He...')
        hhe_cool = self.get_cooling_for_species('HHe', T, nH, fHe)

        if verbose >= environment.VERBOSE_NORMAL:
            print('interpolate ne/nH and solar metallicity ne/nH...')
        hhe_ne = self._interpolate_fHe(self._ne_nH_interp, T, nH, fHe)
        solar_ne = np.array([
            self._solar_ne_nH_interp(t, n) for t, n in zip(T, nH)
        ]).reshape(len(T))

        if verbose >= environment.VERBOSE_NORMAL:
            print('combine to netto cooling rates...')
        net_cool = UnitArr(hhe_cool + (hhe_ne / solar_ne) * metal_cool,
                           'erg cm**3 s**-1')

        # add Compton cooling of the CMB
        if Compton_cooling:  # and self._redshift > 0:
            if verbose >= environment.VERBOSE_NORMAL:
                print('add the Compton cooling rates...')

            from . import quantities
            Stefan = quantities.constants['Stefan-Boltzmann constant']
            c = quantities.c
            m_e = quantities.m_e
            Thompson = quantities.constants['Thomson cross section']
            kB = quantities.kB

            T_CMB = UnitArr('2.728 K') * (1.0 + self._redshift)  # CMB temperature
            # note: nH was made unitless (for speed); [hhe_]ne is unitless anyway
            Compton_cool = -(16.0 * Stefan * Thompson * (T_CMB ** 4) / (m_e * c ** 2)) \
                           * kB * (T_CMB - T) * hhe_ne / UnitArr(nH, 'cm**-3')
            Compton_cool.convert_to('erg cm**3 s**-1')
            # print 'Compton cooling [0%,25%,100%]-perc.:', \
            #        np.percentile(Compton_cool, [0,50,100] ), Compton_cool.units
            # print 'rel. Compton cooling [0%,25%,100%]-perc.:', \
            #        np.percentile((Compton_cool/net_cool).in_units_of(1), [0,50,100] )

            net_cool += Compton_cool

        net_cool.convert_to(units)

        return net_cool




'''
Module for physics that are not directly connected to snapshots.

Also doctest other parts of this sub-module:
    >>> import doctest
    >>> doctest.testmod(quantities)
    TestResults(failed=0, attempted=33)
    >>> doctest.testmod(cosmology)
    TestResults(failed=0, attempted=42)
    >>> doctest.testmod(cooling)
    TestResults(failed=0, attempted=14)
'''

from .quantities import *
from .cosmology import *
from .cooling import *

